import datetime
import glob
import os.path
import re
import shutil
import stat
import threading
import time
import traceback
from sqlite3 import OperationalError
from weakref import WeakKeyDictionary
from xml.etree import ElementTree

import imdb
from unidecode import unidecode
from urllib3.exceptions import MaxRetryError, NewConnectionError

import sickchill
import sickchill.oldbeard.providers
import sickchill.oldbeard.scene_numbering
from sickchill import logger, settings
from sickchill.helper.common import dateTimeFormat, episode_num, remove_extension, replace_extension, sanitize_filename, try_int
from sickchill.helper.exceptions import (
    EpisodeDeletedException,
    EpisodeNotFoundException,
    MultipleEpisodesInDatabaseException,
    MultipleShowObjectsException,
    MultipleShowsInDatabaseException,
    NoNFOException,
    ShowDirectoryNotFoundException,
    ShowNotFoundException,
)
from sickchill.oldbeard import db, helpers, network_timezones, notifiers, postProcessor, subtitles
from sickchill.oldbeard.blackandwhitelist import BlackAndWhiteList
from sickchill.oldbeard.common import (
    ARCHIVED,
    DOWNLOADED,
    FAILED,
    IGNORED,
    NAMING_DUPLICATE,
    NAMING_EXTEND,
    NAMING_LIMITED_EXTEND,
    NAMING_LIMITED_EXTEND_E_PREFIXED,
    NAMING_SEPARATED_REPEAT,
    Overview,
    Quality,
    SKIPPED,
    SNATCHED,
    SNATCHED_PROPER,
    statusStrings,
    UNAIRED,
    UNKNOWN,
    WANTED,
)
from sickchill.oldbeard.name_parser.parser import InvalidNameException, InvalidShowException, NameParser
from sickchill.show.Show import Show

try:
    from send2trash import send2trash
except ImportError:

    def send2trash(path):
        if os.path.isfile(path):
            os.remove(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)


class DirtySetter(object):
    """A descriptor that forbids negative values"""

    def __init__(self, default):
        self.default = default
        self.data = WeakKeyDictionary()

    def __get__(self, instance, owner):
        try:
            return self.data.get(instance, self.default)
        except Exception:
            return None

    def __set__(self, instance, value):
        if value != self.data.get(instance):
            instance.dirty = True

        self.data[instance] = value


class TVShow(object):
    indexerid = DirtySetter(0)
    indexer = DirtySetter(0)
    show_name = DirtySetter("")
    imdb_id = DirtySetter("")
    network = DirtySetter("")
    genre: list = DirtySetter([])
    classification = DirtySetter("")
    runtime = DirtySetter(0)
    imdb_info = DirtySetter({})
    quality = DirtySetter(int(settings.QUALITY_DEFAULT))
    season_folders = DirtySetter(int(settings.SEASON_FOLDERS_DEFAULT))
    status = DirtySetter("Unknown")
    airs = DirtySetter("")
    startyear = DirtySetter(0)
    paused = DirtySetter(0)
    air_by_date = DirtySetter(0)
    subtitles = DirtySetter(int(settings.SUBTITLES_DEFAULT))
    subtitles_sr_metadata = DirtySetter(0)
    dvdorder = DirtySetter(0)
    lang = DirtySetter("en")
    last_update_indexer = DirtySetter(1)
    sports = DirtySetter(0)
    anime = DirtySetter(0)
    scene = DirtySetter(0)
    rls_ignore_words = DirtySetter("")
    rls_require_words = DirtySetter("")
    rls_prefer_words = DirtySetter("")
    default_ep_status = DirtySetter(SKIPPED)
    custom_name = DirtySetter("")

    # location = DirtySetter("")

    def __init__(self, indexer, indexerid: int, lang=""):
        self.dirty = True

        self.lock = threading.Lock()
        self.episodes = {}
        self.nextaired = ""
        self.release_groups = None
        self._location = ""
        self.indexer = indexer
        self.indexerid = indexerid
        self.lang = lang

        otherShow = Show.find(settings.showList, self.indexerid)
        if otherShow is not None:
            raise MultipleShowObjectsException("Can't create a show if it already exists")

        self.loadFromDB()

    @property
    def name(self):
        return self.custom_name or self.show_name

    @name.setter
    def name(self, name):
        self.show_name = name

    @property
    def is_anime(self):
        return int(self.anime) > 0

    @property
    def is_sports(self):
        return int(self.sports) > 0

    @property
    def is_scene(self):
        return int(self.scene) > 0

    @property
    def network_logo_name(self):
        return unidecode(self.network).lower()

    @property
    def sort_name(self):
        return helpers.sortable_name(self.name)

    @property
    def idxr(self):
        return sickchill.indexer[self.indexer]

    @property
    def indexer_name(self):
        return self.idxr.name

    @property
    def network_image_url(self):
        return "images/network/{0}.png".format(unidecode(self.network or "nonetwork").lower())

    def show_image_url(self, which):
        return settings.IMAGE_CACHE.image_url(self.indexerid, which)

    def _getLocation(self):
        # no dir check needed if missing show dirs are created during post-processing
        if settings.CREATE_MISSING_SHOW_DIRS or os.path.isdir(self._location):
            return self._location

        raise ShowDirectoryNotFoundException("Show folder doesn't exist, you shouldn't be using it")

    def _setLocation(self, newLocation):
        logger.debug(f"Setter sets location to {newLocation}")
        # Don't validate dir if user wants to add shows without creating a dir
        if settings.ADD_SHOWS_WO_DIR or os.path.isdir(newLocation):
            if self._location != newLocation:
                self.dirty = True
            self._location = newLocation
        else:
            raise NoNFOException("Invalid folder for the show!")

    location = property(_getLocation, _setLocation)

    # delete references to anything that's not in the internal lists
    def flushEpisodes(self):

        for curSeason in self.episodes:
            self.episodes[curSeason].clear()

        self.episodes.clear()

    def getAllEpisodes(self, season=None, has_location=False):

        sql_selection = "SELECT season, episode, "

        # subselection to detect multi-episodes early, share_location > 0
        sql_selection += "(SELECT COUNT (*) FROM tv_episodes WHERE showid = tve.showid "
        sql_selection += "AND season = tve.season AND location != '' AND location = tve.location "
        sql_selection += "AND episode != tve.episode) AS share_location "
        sql_selection += "FROM tv_episodes tve WHERE showid = {0} ".format(self.indexerid)

        if season is not None:
            sql_selection += "AND season = {0} ".format(season)

        if has_location:
            sql_selection += "AND location != '' "

        # need ORDER episode ASC to rename multi-episodes in order S01E01-02
        sql_selection += "ORDER BY season ASC, episode ASC "

        main_db_con = db.DBConnection()
        results = main_db_con.select(sql_selection)

        ep_list = []
        for cur_result in results:
            cur_ep = self.getEpisode(cur_result["season"], cur_result["episode"])
            if not cur_ep:
                continue

            cur_ep.relatedEps = []
            if cur_ep.location:
                # if there is a location, check if it's a multi-episode (share_location > 0) and put them in relatedEps
                if cur_result["share_location"] > 0:
                    related_eps_result = main_db_con.select(
                        "SELECT season, episode FROM tv_episodes WHERE showid = ? AND season = ? AND location = ? AND episode != ? ORDER BY episode",
                        [self.indexerid, cur_ep.season, cur_ep.location, cur_ep.episode],
                    )
                    for cur_related_ep in related_eps_result:
                        related_ep = self.getEpisode(cur_related_ep["season"], cur_related_ep["episode"])
                        if related_ep and related_ep not in cur_ep.relatedEps:
                            cur_ep.relatedEps.append(related_ep)
            ep_list.append(cur_ep)

        return ep_list

    def getEpisode(self, season=None, episode=None, ep_file=None, noCreate=False, absolute_number=None):
        season = try_int(season, None)
        episode = try_int(episode, None)
        absolute_number = try_int(absolute_number, None)

        if season in self.episodes and episode in self.episodes[season] and self.episodes[season][episode]:
            return self.episodes[season][episode]

        # if we get an anime get the real season and episode
        if self.is_anime and absolute_number and not season and not episode:
            main_db_con = db.DBConnection()
            sql = "SELECT season, episode FROM tv_episodes WHERE showid = ? AND absolute_number = ? AND season != 0"
            sql_results = main_db_con.select(sql, [self.indexerid, absolute_number])

            if len(sql_results) == 1:
                episode = int(sql_results[0]["episode"])
                season = int(sql_results[0]["season"])
                logger.debug("Found episode by absolute number {absolute} which is {ep}".format(absolute=absolute_number, ep=episode_num(season, episode)))
            elif len(sql_results) > 1:
                logger.exception("Multiple entries for absolute number: {absolute} in show: {name} found ".format(absolute=absolute_number, name=self.name))

                return None
            else:
                logger.debug(f"No entries for absolute number: {absolute_number} in show: {self.name} found.")
                return None

        if season not in self.episodes:
            self.episodes[season] = {}

        if not self.episodes[season].get(episode):
            if noCreate:
                return None

            # logger.debug("{id}: An object for episode {ep} didn't exist in the cache, trying to create it".format
            #            (id=self.indexerid, ep=episode_num(season, episode)))

            if ep_file:
                ep = TVEpisode(self, season, episode, ep_file)
            else:
                ep = TVEpisode(self, season, episode)

            if ep:
                self.episodes[season][episode] = ep

        return self.episodes[season][episode]

    def should_update(self, update_date=datetime.date.today()):

        # if show is not 'Ended' always update (status 'Continuing')
        if self.status == "Continuing":
            return True

        # run logic against the current show latest aired and next unaired data to see if we should bypass 'Ended' status

        graceperiod = datetime.timedelta(days=30)

        last_airdate = datetime.date.min

        # get latest aired episode to compare against today - graceperiod and today + graceperiod
        main_db_con = db.DBConnection()
        sql_result = main_db_con.select(
            "SELECT IFNULL(MAX(airdate), 0) as last_aired FROM tv_episodes WHERE showid = ? AND season > 0 AND airdate > 1 AND status > 1", [self.indexerid]
        )

        if sql_result and sql_result[0]["last_aired"] != 0:
            last_airdate = datetime.date.fromordinal(sql_result[0]["last_aired"])
            if (update_date - graceperiod) <= last_airdate <= (update_date + graceperiod):
                return True

        # get next upcoming UNAIRED episode to compare against today + graceperiod
        sql_result = main_db_con.select(
            "SELECT IFNULL(MIN(airdate), 0) as airing_next FROM tv_episodes WHERE showid = ? AND season > 0 AND airdate > 1 AND status = 1", [self.indexerid]
        )

        if sql_result and sql_result[0]["airing_next"] != 0:
            next_airdate = datetime.date.fromordinal(sql_result[0]["airing_next"])
            if next_airdate <= (update_date + graceperiod):
                return True

        last_update_indexer = datetime.date.fromordinal(self.last_update_indexer)

        # in the first year after ended (last airdate), update every 30 days
        if (update_date - last_airdate) < datetime.timedelta(days=450) and (update_date - last_update_indexer) > datetime.timedelta(days=30):
            return True

        return False

    def writeShowNFO(self):

        result = False

        if not os.path.isdir(self._location):
            logger.info(f"{self.indexerid}: Show dir doesn't exist, skipping NFO generation")
            return False

        logger.debug(f"{self.indexerid}: Writing NFOs for show")
        for cur_provider in settings.metadata_provider_dict.values():
            result = cur_provider.create_show_metadata(self) or result

        return result

    def writeMetadata(self, show_only=False):

        if not os.path.isdir(self._location):
            logger.info(f"{self.indexerid}: Show dir doesn't exist, skipping NFO generation")
            return

        self.getImages()

        self.writeShowNFO()

        if not show_only:
            self.writeEpisodeNFOs()

    def writeEpisodeNFOs(self):

        if not os.path.isdir(self._location):
            logger.info(f"{self.indexerid}: Show dir doesn't exist, skipping NFO generation")
            return

        logger.debug(f"{self.indexerid}: Writing NFOs for all episodes")

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT season, episode FROM tv_episodes WHERE showid = ? AND location != ''", [self.indexerid])

        for epResult in sql_results:
            logger.debug("{id}: Retrieving/creating episode {ep}".format(id=self.indexerid, ep=episode_num(epResult["season"], epResult["episode"])))
            curEp = self.getEpisode(epResult["season"], epResult["episode"])
            if not curEp:
                continue

            curEp.createMetaFiles()

    def updateMetadata(self):

        if not os.path.isdir(self._location):
            logger.info(f"{self.indexerid}: Show dir doesn't exist, skipping NFO generation")
            return

        result = False

        if not os.path.isdir(self._location):
            logger.info(f"{self.indexerid}: Show dir doesn't exist, skipping NFO generation")
            return False

        logger.info(f"{self.indexerid}: Updating NFOs for show with new indexer info")
        for cur_provider in settings.metadata_provider_dict.values():
            result = cur_provider.update_show_indexer_metadata(self) or result

        return result

    # find all media files in the show folder and create episodes for as many as possible
    def loadEpisodesFromDir(self):

        if not os.path.isdir(self._location):
            logger.debug(f"{self.indexerid}: Show dir doesn't exist, not loading episodes from disk")
            return

        logger.debug(f"{self.indexerid}: Loading all episodes from the show directory {self._location}")

        # get file list
        media_files = helpers.list_media_files(self._location)

        # create TVEpisodes from each media file (if possible)
        sql_l = []
        for media_file in media_files:
            logger.debug("{tvdbid}: Creating episode from {filename}".format(tvdbid=str(self.indexerid), filename=os.path.basename(media_file)))
            curEpisode = None
            try:
                curEpisode = self.makeEpFromFile(media_file)
            except (ShowNotFoundException, EpisodeNotFoundException) as error:
                logger.exception("Episode {filename} returned an exception: {ex}".format(filename=os.path.basename(media_file), ex=str(error)))
                continue
            except EpisodeDeletedException:
                logger.debug("The episode deleted itself when I tried making an object for it")

            if curEpisode is None:
                continue

            # see if we should save the release name in the db
            ep_filename = os.path.basename(curEpisode.location)
            ep_filename = os.path.splitext(ep_filename)[0]

            try:
                parse_result = NameParser(False, showObj=self, tryIndexers=True).parse(ep_filename)
            except (InvalidNameException, InvalidShowException):
                parse_result = None

            if " " not in ep_filename and parse_result and parse_result.release_group:
                logger.debug(f"Name {ep_filename} gave release group of {parse_result.release_group}, seems valid")
                curEpisode.release_name = ep_filename
                curEpisode.release_group = parse_result.release_group

            # store the reference in the show
            if curEpisode is not None:
                if self.subtitles:
                    try:
                        curEpisode.refreshSubtitles()
                    except Exception:
                        logger.exception("{0}: Could not refresh subtitles".format(self.indexerid))
                        logger.debug(traceback.format_exc())

                sql_l.append(curEpisode.get_sql())

        if sql_l:
            main_db_con = db.DBConnection()
            main_db_con.mass_action(sql_l)

    def loadEpisodesFromDB(self):

        logger.debug("Loading all episodes from the database")
        scannedEps = {}

        try:
            main_db_con = db.DBConnection()
            sql = "SELECT season, episode, showid, show_name FROM tv_episodes JOIN tv_shows WHERE showid = indexer_id and showid = ?"
            sql_results = main_db_con.select(sql, [self.indexerid])
        except Exception as error:
            logger.exception("Could not load episodes from the DB. Error: {0}".format(error))
            return scannedEps

        curShowid = None
        curShowName = None

        for curResult in sql_results:

            curSeason = int(curResult["season"])
            curEpisode = int(curResult["episode"])
            curShowid = int(curResult["showid"])
            curShowName = str(curResult["show_name"])

            if curSeason not in scannedEps:
                logger.debug("{id}: Not curSeason in scannedEps".format(id=curShowid))
                scannedEps[curSeason] = {}

            logger.debug("{id}: Loading {show} {ep} from the DB".format(id=curShowid, show=curShowName, ep=episode_num(curSeason, curEpisode)))

            try:
                curEp = self.getEpisode(curSeason, curEpisode)
                if not curEp:
                    raise EpisodeNotFoundException

                curEp.loadFromDB(curSeason, curEpisode)
                scannedEps[curSeason][curEpisode] = True
            except EpisodeDeletedException:
                logger.debug(
                    "{id}: Tried loading {show} {ep} from the DB that should have been deleted, skipping it".format(
                        id=curShowid, show=curShowName, ep=episode_num(curSeason, curEpisode)
                    )
                )
                continue

        if curShowName and curShowid:
            logger.debug("{id}: Finished loading all episodes for {show} from the DB".format(show=curShowName, id=curShowid))

        return scannedEps

    def loadEpisodesFromIndexer(self):
        logger.debug(_("{show_id}: Loading all episodes from {indexer_name}...").format(show_id=self.indexerid, indexer_name=self.indexer_name))

        scannedEps = {}

        sql_l = []
        for series_episode in self.idxr.episodes(self):
            if series_episode["airedSeason"] not in scannedEps:
                scannedEps[series_episode["airedSeason"]] = {}

            season_episode = series_episode["airedSeason"], series_episode["airedEpisodeNumber"]
            try:
                show_episode = self.getEpisode(*season_episode)
                if not show_episode:
                    raise EpisodeNotFoundException
            except EpisodeNotFoundException:
                logger.info(
                    "{id}: {indexer} object for {ep} is incomplete, skipping this episode".format(
                        id=self.indexerid, indexer=self.indexer_name, ep=episode_num(*season_episode)
                    )
                )
                continue
            else:
                try:
                    with show_episode.lock:
                        show_episode.loadFromIndexer(*season_episode)
                        sql_l.append(show_episode.get_sql())
                except EpisodeDeletedException:
                    logger.info("The episode was deleted, skipping the rest of the load")
                    continue

            scannedEps[series_episode["airedSeason"]][series_episode["airedEpisodeNumber"]] = True

        if sql_l:
            main_db_con = db.DBConnection()
            main_db_con.mass_action(sql_l)

        # Done updating save last update date
        self.last_update_indexer = datetime.datetime.now().toordinal()

        self.saveToDB()

        return scannedEps

    def getImages(self):
        fanart_result = poster_result = banner_result = False
        season_posters_result = season_banners_result = season_all_poster_result = season_all_banner_result = False

        for cur_provider in settings.metadata_provider_dict.values():

            # logger.debug(f"Running metadata routines for {cur_provider.name}")

            fanart_result = cur_provider.create_fanart(self) or fanart_result
            poster_result = cur_provider.create_poster(self) or poster_result
            banner_result = cur_provider.create_banner(self) or banner_result

            season_posters_result = cur_provider.create_season_posters(self) or season_posters_result
            season_banners_result = cur_provider.create_season_banners(self) or season_banners_result
            season_all_poster_result = cur_provider.create_season_all_poster(self) or season_all_poster_result
            season_all_banner_result = cur_provider.create_season_all_banner(self) or season_all_banner_result

        return (
            fanart_result
            or poster_result
            or banner_result
            or season_posters_result
            or season_banners_result
            or season_all_poster_result
            or season_all_banner_result
        )

    # make a TVEpisode object from a media file
    def makeEpFromFile(self, filepath):

        if not os.path.isfile(filepath):
            logger.info("{0}: That isn't even a real file dude... {1}".format(self.indexerid, filepath))
            return None

        logger.debug("{0}: Creating episode object from {1}".format(self.indexerid, filepath))

        try:
            parse_result = NameParser(showObj=self, tryIndexers=True, parse_method=("normal", "anime")[self.is_anime]).parse(filepath, True, True)
        except (InvalidNameException, InvalidShowException) as error:
            logger.debug("{0}: {1}".format(self.indexerid, error))
            return None

        episodes = [ep for ep in parse_result.episode_numbers if ep is not None]
        if not episodes:
            logger.info("{0}: parse_result: {1}".format(self.indexerid, parse_result))
            logger.info("{0}: No episode number found in {1}, ignoring it".format(self.indexerid, filepath))
            return None

        # for now lets assume that any episode in the show dir belongs to that show
        season = parse_result.season_number if parse_result.season_number is not None else 1
        rootEp = None

        sql_l = []
        for current_ep in episodes:
            logger.debug("{0}: {1} parsed to {2} {3}".format(self.indexerid, filepath, self.name, episode_num(season, current_ep)))

            checkQualityAgain = False
            same_file = False

            curEp = self.getEpisode(season, current_ep)
            if not curEp:
                try:
                    curEp = self.getEpisode(season, current_ep, filepath)
                    if not curEp:
                        raise EpisodeNotFoundException
                except EpisodeNotFoundException:
                    logger.exception("{0}: Unable to figure out what this file is, skipping {1}".format(self.indexerid, filepath))
                    continue

            else:
                # if there is a new file associated with this ep then re-check the quality
                if curEp.location and os.path.normpath(curEp.location) != os.path.normpath(filepath):
                    logger.debug(
                        "{0}: The old episode had a different file associated with it, re-checking the quality using the new filename {1}".format(
                            self.indexerid, filepath
                        )
                    )
                    checkQualityAgain = True

                with curEp.lock:
                    old_size = curEp.file_size
                    curEp.location = filepath
                    # if the sizes are the same then it's probably the same file
                    same_file = old_size and curEp.file_size == old_size
                    curEp.checkForMetaFiles()

            if rootEp is None:
                rootEp = curEp
            else:
                if curEp not in rootEp.relatedEps:
                    with rootEp.lock:
                        rootEp.relatedEps.append(curEp)

            # if it's a new file then
            if not same_file:
                with curEp.lock:
                    curEp.release_name = ""
                    curEp.release_group = ""

            # if they replace a file on me I'll make some attempt at re-checking the quality unless I know it's the same file
            if checkQualityAgain and not same_file:
                newQuality = Quality.nameQuality(filepath, self.is_anime)
                logger.debug(
                    "{0}: Since this file has been renamed, I checked {1} and found quality {2}".format(
                        self.indexerid, filepath, Quality.qualityStrings[newQuality]
                    )
                )

                with curEp.lock:
                    curEp.status = Quality.compositeStatus(DOWNLOADED, newQuality)

            # check for status/quality changes as long as it's a new file
            elif not same_file and sickchill.oldbeard.helpers.is_media_file(filepath) and curEp.status not in Quality.DOWNLOADED + Quality.ARCHIVED + [IGNORED]:
                oldStatus, oldQuality = Quality.splitCompositeStatus(curEp.status)
                newQuality = Quality.nameQuality(filepath, self.is_anime)

                newStatus = None

                # if it was snatched and now exists then set the status correctly
                if oldStatus == SNATCHED and oldQuality <= newQuality:
                    logger.debug(
                        "{0}: This ep used to be snatched with quality {1} but a file exists with quality {2} so I'm setting the status to DOWNLOADED".format(
                            self.indexerid, Quality.qualityStrings[oldQuality], Quality.qualityStrings[newQuality]
                        )
                    )
                    newStatus = DOWNLOADED

                # if it was snatched proper and we found a higher quality one then allow the status change
                elif oldStatus == SNATCHED_PROPER and oldQuality < newQuality:
                    logger.debug(
                        "{0}: This ep used to be snatched proper with quality {1} but a file exists with quality {2} so I'm setting the status to DOWNLOADED".format(
                            self.indexerid, Quality.qualityStrings[oldQuality], Quality.qualityStrings[newQuality]
                        )
                    )
                    newStatus = DOWNLOADED

                elif oldStatus not in (SNATCHED, SNATCHED_PROPER):
                    newStatus = DOWNLOADED

                if newStatus is not None:
                    with curEp.lock:
                        logger.debug(
                            "{0}: We have an associated file, so setting the status from {1} to DOWNLOADED/{2}".format(
                                self.indexerid, curEp.status, Quality.statusFromName(filepath, anime=self.is_anime)
                            )
                        )
                        curEp.status = Quality.compositeStatus(newStatus, newQuality)

            with curEp.lock:
                sql_l.append(curEp.get_sql())

        if sql_l:
            main_db_con = db.DBConnection()
            main_db_con.mass_action(sql_l)

        # creating metafiles on the root should be good enough
        if rootEp:
            with rootEp.lock:
                rootEp.createMetaFiles()

        return rootEp

    def loadFromDB(self):

        # logger.debug(f"{self.indexerid}: Loading show info from database")

        main_db_con = db.DBConnection(row_type="dict")
        sql_results = main_db_con.select("SELECT * FROM tv_shows WHERE indexer_id = ?", [self.indexerid])

        if len(sql_results) > 1:
            raise MultipleShowsInDatabaseException()
        elif not sql_results:
            # logger.info(f"{self.indexerid}: Unable to find the show in the database")
            return
        else:
            self.indexer = int(sql_results[0]["indexer"] or 0)

            if not self.show_name:
                self.name = sql_results[0]["show_name"]
            if not self.custom_name:
                self.custom_name = sql_results[0]["custom_name"]
            if not self.network:
                self.network = sql_results[0]["network"]
            if not self.genre:
                self.genre = [x.strip() for x in sql_results[0]["genre"].split("|") if x.strip()]

            if not self.classification:
                self.classification = sql_results[0]["classification"]

            self.runtime = sql_results[0]["runtime"]

            self.status = sql_results[0]["status"]
            if self.status is None:
                self.status = "Unknown"

            self.airs = sql_results[0]["airs"]
            if self.airs is None:
                self.airs = ""

            self.startyear = int(sql_results[0]["startyear"] or 0)
            self.air_by_date = int(sql_results[0]["air_by_date"] or 0)
            self.anime = int(sql_results[0]["anime"] or 0)
            self.sports = int(sql_results[0]["sports"] or 0)
            self.scene = int(sql_results[0]["scene"] or 0)
            self.subtitles = int(sql_results[0]["subtitles"] or 0)
            self.dvdorder = int(sql_results[0]["dvdorder"] or 0)
            self.quality = int(sql_results[0]["quality"] or UNKNOWN)
            self.season_folders = int(not int(sql_results[0]["flatten_folders"] or 0))  # FIXME: inverted until next database version
            self.paused = int(sql_results[0]["paused"] or 0)
            self.paused = int(sql_results[0]["paused"] or 0)

            self._location = sql_results[0]["location"]

            if not self.lang:
                self.lang = sql_results[0]["lang"]

            self.last_update_indexer = sql_results[0]["last_update_indexer"]

            self.rls_ignore_words = sql_results[0]["rls_ignore_words"]
            self.rls_require_words = sql_results[0]["rls_require_words"]
            self.rls_prefer_words = sql_results[0]["rls_prefer_words"]

            self.default_ep_status = int(sql_results[0]["default_ep_status"] or SKIPPED)

            if not self.imdb_id:
                self.imdb_id = sql_results[0]["imdb_id"]

            if self.is_anime:
                self.release_groups = BlackAndWhiteList(self.indexerid)

            self.subtitles_sr_metadata = int(sql_results[0]["sub_use_sr_metadata"] or 0)

        # Get IMDb_info from database
        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT * FROM imdb_info WHERE indexer_id = ?", [self.indexerid])

        if not sql_results:
            self.load_imdb_imfo()
            sql_results = main_db_con.select("SELECT * FROM imdb_info WHERE indexer_id = ?", [self.indexerid])
            if not sql_results:
                logger.info(f"{self.indexerid}: Unable to find IMDb show info in the database")
                return

        self.imdb_info = dict(zip(sql_results[0].keys(), sql_results[0]))
        self.dirty = False
        return True

    def loadFromIndexer(self):
        logger.debug(f"{self.indexerid}: Loading show info from {self.indexer_name}")

        myShow = sickchill.indexer.series(self)
        if not myShow or not getattr(myShow, "seriesName"):
            raise AttributeError("Found {0}, but attribute 'seriesName' was empty.".format(self.indexerid))

        self.name = myShow.seriesName.strip()
        self.classification = getattr(myShow, "classification", "Scripted")
        self.genre = getattr(myShow, "genre", [])
        self.network = getattr(myShow, "network", "")
        self.runtime = getattr(myShow, "runtime", "")

        self.imdb_id = getattr(myShow, "imdbId", "")

        if hasattr(myShow, "airsDayOfWeek") and hasattr(myShow, "airsTime"):
            if myShow.airsTime and myShow.airsDayOfWeek:
                self.airs = f"{myShow.airsDayOfWeek} {myShow.airsTime}".strip()
            else:
                self.airs = self.airs.strip() if self.airs else ""

        if self.airs is None:
            self.airs = ""

        if hasattr(myShow, "firstAired") and myShow.firstAired:
            self.startyear = int(myShow.firstAired.split("-")[0])

        self.status = getattr(myShow, "status", "Unknown")

    def check_imdb_id(self):
        if self.imdb_id:
            self.imdb_id = re.sub(r"[^\d]", "", self.imdb_id)

        if self.imdb_id:
            self.imdb_id = "tt" + self.imdb_id

        try:
            int(self.imdb_id.lstrip("t"))
        except (ValueError, TypeError):
            self.imdb_id = ""

    def load_imdb_imfo(self):
        try:
            client = imdb.IMDb()

            # Check that the imdb_id we have is valid for searching
            self.check_imdb_id()

            if self.name and not self.imdb_id:
                # Add regular name and custom name to be searched first
                attempts = {self.show_name}
                if self.custom_name:
                    attempts.add(self.custom_name)

                # Now the generated name/year search string types for normal name
                attempts.update({f"{self.show_name} ({self.startyear})", f'"{self.show_name}" ({self.startyear})'})

                if self.custom_name:
                    # Now the generated name/year search string types for custom name
                    attempts.update({f"{self.custom_name} ({self.startyear})", f'"{self.custom_name}" ({self.startyear})'})

                for attempt in attempts:
                    results = [x for x in client.search_movie(attempt) if x["kind"] == "tv series" and x["title"] == attempt]
                    if self.startyear:
                        results = [x for x in results if x["year"] == self.startyear]
                    if len(results) == 1:
                        self.imdb_id = results[0]["imdbID"]

                    if self.imdb_id:
                        break

            # Make sure the lib didn't give us back something bogus
            self.check_imdb_id()

            if not self.imdb_id:
                logger.debug(f"{self.indexerid}: Not loading show info from IMDb, because we don't know the imdb_id")
                return

            logger.debug(f"{self.indexerid}: Loading show info from IMDb")
            imdb_title: dict = client.get_movie(self.imdb_id.strip("t"))

            self.imdb_info = {
                "indexer_id": self.indexerid,
                "imdb_id": imdb_title.setdefault("imdbID", self.imdb_id),
                "title": imdb_title.setdefault("title", self.name),
                "year": imdb_title.setdefault("year", self.startyear),
                "akas": "|".join(imdb_title.setdefault("akas", [])),
                "runtimes": imdb_title.setdefault("runtimes", [self.runtime])[0],
                "genres": "|".join(imdb_title.setdefault("genres", [])),
                "countries": "|".join(imdb_title.get("countries", [])),
                "country_codes": "|".join(imdb_title.get("country codes", [])),
                "certificates": "|".join(imdb_title.setdefault("certificates", [])),
                "rating": str(imdb_title.setdefault("rating", 0.0)),
                "votes": str(imdb_title.setdefault("votes", 0)),
                "last_update": datetime.date.today().toordinal(),
            }

            logger.debug(f"{self.indexerid}: Obtained info from IMDb ->{self.imdb_info}")
        except (ValueError, LookupError, OperationalError, imdb.IMDbError, NewConnectionError, MaxRetryError) as error:
            logger.info(f"Could not get IMDB info: {error}")
        except (SyntaxError, KeyError):
            logger.info("Could not get info from IDMb, pip install lxml")

    def nextEpisode(self):
        curDate = datetime.date.today().toordinal()
        if not self.nextaired or self.nextaired and curDate > self.nextaired:
            main_db_con = db.DBConnection()
            sql_results = main_db_con.select(
                "SELECT airdate, season, episode FROM tv_episodes WHERE showid = ? AND airdate >= ? AND status IN (?,?) ORDER BY airdate LIMIT 1",
                [self.indexerid, datetime.date.today().toordinal(), UNAIRED, WANTED],
            )

            self.nextaired = sql_results[0]["airdate"] if sql_results else ""

        return self.nextaired

    def deleteShow(self, full=False):
        main_db_con = db.DBConnection()

        episodes_locations = main_db_con.select("SELECT location FROM tv_episodes WHERE showid = ? AND location != ''", [self.indexerid])

        sql_l = [
            ["DELETE FROM tv_episodes WHERE showid = ?", [self.indexerid]],
            ["DELETE FROM tv_shows WHERE indexer_id = ?", [self.indexerid]],
            ["DELETE FROM imdb_info WHERE indexer_id = ?", [self.indexerid]],
            ["DELETE FROM xem_refresh WHERE indexer_id = ?", [self.indexerid]],
            ["DELETE FROM scene_numbering WHERE indexer_id = ?", [self.indexerid]],
            ["DELETE FROM history WHERE showid = ?", [self.indexerid]],
            ["DELETE FROM indexer_mapping WHERE indexer_id = ?", [self.indexerid]],
            ["DELETE FROM blacklist WHERE show_id = ?", [self.indexerid]],
            ["DELETE FROM whitelist WHERE show_id = ?", [self.indexerid]],
        ]

        main_db_con.mass_action(sql_l)

        cache_db_con = db.DBConnection("cache.db")
        sql_l = [["DELETE FROM scene_exceptions WHERE indexer_id = ?", [self.indexerid]], ["DELETE FROM scene_names WHERE indexer_id = ?", [self.indexerid]]]

        cache_db_con.mass_action(sql_l)

        for provider in sickchill.oldbeard.providers.__all__:
            if cache_db_con.has_table(provider) and cache_db_con.has_column(provider, "indexerid"):
                # language=TEXT
                cache_db_con.action("delete from {} WHERE indexerid = ?".format(provider), [self.indexerid])

        failed_db_con = db.DBConnection("failed.db")
        sql_l = [
            ["DELETE FROM history WHERE showid = ?", [self.indexerid]],
        ]

        failed_db_con.mass_action(sql_l)

        action = ("delete", "trash")[settings.TRASH_REMOVE_SHOW]

        # remove self from show list
        settings.showList = [x for x in settings.showList if int(x.indexerid) != self.indexerid]

        # clear the cache
        image_cache_dir = os.path.join(settings.CACHE_DIR, "images")
        for cache_file in glob.glob(os.path.join(glob.escape(image_cache_dir), f"{self.indexerid}.*")):
            logger.info("Attempt to {0} cache file {1}".format(action, cache_file))
            try:
                if settings.TRASH_REMOVE_SHOW:
                    send2trash(cache_file)
                else:
                    os.remove(cache_file)

            except OSError as error:
                logger.warning("Unable to {0} {1}: {2}".format(action, cache_file, error))

        # remove entire show folder
        if full:
            try:
                logger.info("Attempt to {0} show folder {1}".format(action, self._location))
                # check first the read-only attribute
                file_attribute = os.stat(self.location)[0]
                if not file_attribute & stat.S_IWRITE:
                    # File is read-only, so make it writeable
                    logger.debug("Attempting to make writeable the read only folder {0}".format(self._location))
                    try:
                        os.chmod(self.location, stat.S_IWRITE)
                    except Exception as error:
                        logger.warning("Unable to change permissions of {0}: {1}".format(self._location, error))

                shows_in_folder = main_db_con.select(
                    "SELECT location from tv_shows WHERE location LIKE ? AND indexer_id != ?", ["{}%".format(self.location), self.indexerid]
                )

                num_shows_in_folder = len(shows_in_folder)
                if num_shows_in_folder:
                    logger.info(
                        "Cannot delete the show folder from disk, because this location is the root dir for {num} other shows!".format(num=num_shows_in_folder)
                    )
                    logger.info("Deleting individual episodes. There may be some related files or folders left behind afterwards.")
                    for ep_file in episodes_locations:
                        for show_file in glob.glob(helpers.replace_extension(glob.escape(ep_file["location"]), "*")):
                            logger.info("Attempt to {0} related file {1}".format(action, show_file))
                            try:
                                if settings.TRASH_REMOVE_SHOW:
                                    send2trash(show_file)
                                else:
                                    os.remove(show_file)
                            except OSError as error:
                                logger.warning("Unable to {0} {1}: {2}".format(action, show_file, error))
                else:
                    if settings.TRASH_REMOVE_SHOW:
                        send2trash(self.location)
                    else:
                        shutil.rmtree(self.location)

                    logger.info("{0} show folder {1}".format(("Deleted", "Trashed")[settings.TRASH_REMOVE_SHOW], self._location))

            except ShowDirectoryNotFoundException:
                logger.warning("Show folder does not exist, no need to {0} {1}".format(action, self._location))
            except OSError as error:
                logger.warning("Unable to {0} {1}: {2}".format(action, self._location, error))

        if settings.USE_TRAKT and settings.TRAKT_SYNC_WATCHLIST:
            logger.debug(f"Removing show: indexerid {self.indexerid}, Title {self.name} from Watchlist")
            notifiers.trakt_notifier.update_watchlist(self, update="remove")

    def populateCache(self):
        logger.debug(f"Checking & filling cache for show {self.name}")
        settings.IMAGE_CACHE.fill_cache(self)

    def refreshDir(self):

        # make sure the show dir is where we think it is unless dirs are created on the fly
        if not os.path.isdir(self._location) and not settings.CREATE_MISSING_SHOW_DIRS:
            logger.info(
                "Show dir does not exist, and `create missing show dirs` is disabled. Skipping refresh (statuses will not be updated): {}".format(
                    self._location
                )
            )
            return False

        # load from dir
        self.loadEpisodesFromDir()

        # run through all locations from DB, check that they exist
        logger.debug(f"{self.indexerid}: Loading all episodes with a location from the database")

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT season, episode, location FROM tv_episodes WHERE showid = ? AND location != ''", [self.indexerid])

        sql_l = []
        for ep in sql_results:
            curLoc = os.path.normpath(ep["location"])
            season = int(ep["season"])
            episode = int(ep["episode"])

            try:
                curEp = self.getEpisode(season, episode)
                if not curEp:
                    raise EpisodeDeletedException
            except EpisodeDeletedException:
                logger.debug("The episode was deleted while we were refreshing it, moving on to the next one")
                continue

            # if the path doesn't exist or if it's not in our show dir
            if not os.path.isfile(curLoc) or not os.path.normpath(curLoc).startswith(os.path.normpath(self._location)):

                # check if downloaded files still exist, update our data if this has changed
                if not settings.SKIP_REMOVED_FILES:
                    with curEp.lock:
                        # if it used to have a file associated with it and it doesn't anymore then set it to oldbeard.EP_DEFAULT_DELETED_STATUS
                        if curEp.status in Quality.DOWNLOADED:

                            if settings.EP_DEFAULT_DELETED_STATUS == ARCHIVED:
                                oldStatus_, oldQuality = Quality.splitCompositeStatus(curEp.status)
                                new_status = Quality.compositeStatus(ARCHIVED, oldQuality)
                            else:
                                new_status = settings.EP_DEFAULT_DELETED_STATUS

                            logger.debug(
                                "{id}: Location for {ep} doesn't exist, removing it and changing our status to {status}".format(
                                    id=self.indexerid, ep=episode_num(season, episode), status=statusStrings[new_status]
                                )
                            )
                            curEp.status = new_status
                            curEp.subtitles = list()
                            curEp.subtitles_searchcount = 0
                            curEp.subtitles_lastsearch = str(datetime.datetime.min)
                        curEp.location = ""
                        curEp.hasnfo = False
                        curEp.hastbn = False
                        curEp.release_name = ""
                        curEp.release_group = ""

                        sql_l.append(curEp.get_sql())
                else:
                    logger.info("Skipping updating removed file because `skip removed files` is enabled: {}".format(curLoc))

        if sql_l:
            main_db_con = db.DBConnection()
            main_db_con.mass_action(sql_l)

    def download_subtitles(self, force=False):
        if not os.path.isdir(self._location):
            logger.debug(f"{self.indexerid}: Show dir doesn't exist, can't download subtitles")
            return

        logger.debug("{0}: Downloading subtitles".format(self.indexerid))

        try:
            episodes = self.getAllEpisodes(has_location=True)
            if not episodes:
                logger.debug("{0}: No episodes to download subtitles for {1}".format(self.indexerid, self.name))
                return

            for episode in episodes:
                episode.download_subtitles(force=force)

        except Exception:
            logger.debug("{0}: Error occurred when downloading subtitles for {1}".format(self.indexerid, self.name))
            logger.exception(traceback.format_exc())

    def saveToDB(self, forceSave=False):

        if not self.dirty and not forceSave:
            # logger.debug(f"{self.indexerid}: Not saving show to db - record is not dirty")
            return

        logger.debug("{0:d}: Saving to database: {1}".format(self.indexerid, self.name))

        controlValueDict = {"indexer_id": self.indexerid}
        newValueDict = {
            "indexer": self.indexer,
            "show_name": self.show_name,
            "custom_name": self.custom_name,
            "location": self._location,
            "network": self.network,
            "genre": "|".join(self.genre) if isinstance(self.genre, list) else self.genre,
            "classification": self.classification,
            "runtime": self.runtime,
            "quality": self.quality,
            "airs": self.airs,
            "status": self.status,
            "flatten_folders": int(not self.season_folders),  # FIXME: inverted until next database version
            "paused": self.paused,
            "air_by_date": self.air_by_date,
            "anime": self.anime,
            "scene": self.scene,
            "sports": self.sports,
            "subtitles": self.subtitles,
            "dvdorder": self.dvdorder,
            "startyear": self.startyear,
            "lang": self.lang,
            "imdb_id": self.imdb_id,
            "last_update_indexer": self.last_update_indexer,
            "rls_ignore_words": self.rls_ignore_words,
            "rls_require_words": self.rls_require_words,
            "rls_prefer_words": self.rls_prefer_words,
            "default_ep_status": self.default_ep_status,
            "sub_use_sr_metadata": self.subtitles_sr_metadata,
        }

        main_db_con = db.DBConnection()
        main_db_con.upsert("tv_shows", newValueDict, controlValueDict)

        helpers.update_anime_support()

        if self.imdb_id and self.imdb_info:
            main_db_con = db.DBConnection()
            main_db_con.upsert("imdb_info", self.imdb_info, controlValueDict)

    def __str__(self):
        info_list = [
            f"indexerid: {self.indexerid}",
            f"indexer: {self.indexer}",
            f"name: {self.name}",
            f"location: {self._location}",
            f"status: {self.status}",
            f"startyear: {self.startyear}",
            f"classification: {self.classification}",
            f"runtime: {self.runtime}",
            f"quality: {self.quality}",
            f"scene: {self.is_scene}",
            f"sports: {self.is_sports}",
            f"anime: {self.is_anime}",
        ]
        if self.network:
            info_list.append(f"network: {self.network}")
        if self.airs:
            info_list.append(f"airs: {self.airs}")
        if self.genre:
            info_list.append(f"genre: {'|'.join(self.genre)}")
        return "\n".join(info_list)

    @staticmethod
    def qualitiesToString(qualities=None):
        return ", ".join([Quality.qualityStrings[quality] for quality in qualities or [] if quality and quality in Quality.qualityStrings]) or "None"

    def wantEpisode(self, season, episode, quality, manualSearch=False, downCurQuality=False):

        # if the quality isn't one we want under any circumstances then just say no
        allowed_qualities, preferred_qualities = Quality.splitQuality(self.quality)
        logger.debug(
            "Any,Best = [ {0} ] [ {1} ] Found = [ {2} ]".format(
                self.qualitiesToString(allowed_qualities), self.qualitiesToString(preferred_qualities), self.qualitiesToString([quality])
            )
        )

        if quality not in allowed_qualities + preferred_qualities or quality is UNKNOWN:
            logger.debug(
                "Don't want this quality, ignoring found result for {name} {ep} with quality {quality}".format(
                    name=self.name, ep=episode_num(season, episode), quality=Quality.qualityStrings[quality]
                )
            )
            return False

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT status FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?", [self.indexerid, season, episode])

        if not sql_results or not len(sql_results):
            logger.debug(
                "Unable to find a matching episode in database, ignoring found result for {name} {ep} with quality {quality}".format(
                    name=self.name, ep=episode_num(season, episode), quality=Quality.qualityStrings[quality]
                )
            )
            return False

        epStatus = int(sql_results[0]["status"])
        epStatus_text = statusStrings[epStatus]

        # if we know we don't want it then just say no
        if epStatus in Quality.ARCHIVED + [UNAIRED, SKIPPED, IGNORED] and not manualSearch:
            logger.debug(
                "Existing episode status is '{status}', ignoring found result for {name} {ep} with quality {quality}".format(
                    status=epStatus_text, name=self.name, ep=episode_num(season, episode), quality=Quality.qualityStrings[quality]
                )
            )
            return False

        curStatus_, curQuality = Quality.splitCompositeStatus(epStatus)

        # if it's one of these then we want it as long as it's in our allowed initial qualities
        if epStatus in (WANTED, SKIPPED, UNKNOWN, FAILED):
            logger.debug(
                "Existing episode status is '{status}', getting found result for {name} {ep} with quality {quality}".format(
                    status=epStatus_text, name=self.name, ep=episode_num(season, episode), quality=Quality.qualityStrings[quality]
                )
            )
            return True
        elif manualSearch:
            if (downCurQuality and quality >= curQuality) or (not downCurQuality and quality != curQuality):
                logger.debug(
                    "Usually ignoring found result, but forced search allows the quality,"
                    " getting found result for {name} {ep} with quality {quality}".format(
                        name=self.name, ep=episode_num(season, episode), quality=Quality.qualityStrings[quality]
                    )
                )
                return True

        # if we are re-downloading then we only want it if it's in our preferred_qualities list and better than what we have,
        # or we only have one bestQuality and we do not have that quality yet
        if (
            epStatus in Quality.DOWNLOADED + Quality.SNATCHED + Quality.SNATCHED_PROPER
            and quality in preferred_qualities
            and (quality > curQuality or curQuality not in preferred_qualities)
        ):
            logger.debug(
                "Episode already exists with quality {existing_quality} but the found result"
                " quality {new_quality} is wanted more, getting found result for {name} {ep}".format(
                    existing_quality=Quality.qualityStrings[curQuality],
                    new_quality=Quality.qualityStrings[quality],
                    name=self.name,
                    ep=episode_num(season, episode),
                )
            )
            return True
        elif curQuality == Quality.UNKNOWN and manualSearch:
            logger.debug(
                "Episode already exists but quality is Unknown, getting found result for {name} {ep} with quality {quality}".format(
                    name=self.name, ep=episode_num(season, episode), quality=Quality.qualityStrings[quality]
                )
            )
            return True
        else:
            logger.debug(
                "Episode already exists with quality {existing_quality} and the found result has same/lower quality,"
                " ignoring found result for {name} {ep} with quality {new_quality}".format(
                    existing_quality=Quality.qualityStrings[curQuality],
                    name=self.name,
                    ep=episode_num(season, episode),
                    new_quality=Quality.qualityStrings[quality],
                )
            )
        return False

    def getOverview(self, epStatus, backlog=False):
        """
        Get the Overview status from the Episode status

        :param epStatus: an Episode status
        :return: an Overview status
        """

        ep_status = try_int(epStatus) or UNKNOWN

        if ep_status == WANTED:
            return Overview.WANTED
        elif ep_status in (UNAIRED, UNKNOWN):
            return Overview.UNAIRED
        elif ep_status in (SKIPPED, IGNORED):
            return Overview.SKIPPED
        elif ep_status in Quality.ARCHIVED:
            return Overview.GOOD
        elif ep_status in Quality.FAILED:
            return Overview.WANTED
        elif ep_status in Quality.SNATCHED:
            return Overview.SNATCHED
        elif ep_status in Quality.SNATCHED_PROPER:
            return Overview.SNATCHED_PROPER
        elif ep_status in Quality.SNATCHED_BEST:
            return Overview.SNATCHED_BEST
        elif ep_status in Quality.DOWNLOADED:
            allowed_qualities, preferred_qualities = Quality.splitQuality(self.quality)
            ep_status, cur_quality = Quality.splitCompositeStatus(ep_status)

            if cur_quality not in allowed_qualities + preferred_qualities and not backlog:
                return Overview.QUAL
            elif preferred_qualities and cur_quality not in preferred_qualities and not backlog:
                return Overview.QUAL
            else:
                return Overview.GOOD
        else:
            logger.exception("Could not parse episode status into a valid overview status: {0}".format(epStatus))

    def __getstate__(self):
        d = dict(self.__dict__)
        del d["lock"]
        return d

    def __setstate__(self, d):
        d["lock"] = threading.Lock()
        self.__dict__.update(d)


class TVEpisode(object):
    name = DirtySetter("")
    season = DirtySetter(None)
    episode = DirtySetter(None)
    absolute_number = DirtySetter(0)
    description = DirtySetter("")
    subtitles = DirtySetter(list())
    subtitles_searchcount = DirtySetter(0)
    subtitles_lastsearch = DirtySetter(str(datetime.datetime.min))
    airdate = DirtySetter(datetime.date.min)
    hasnfo = DirtySetter(False)
    hastbn = DirtySetter(False)
    status = DirtySetter(UNKNOWN)
    indexerid = DirtySetter(0)
    file_size = DirtySetter(0)
    release_name = DirtySetter("")
    is_proper = DirtySetter(False)
    version = DirtySetter(0)
    release_group = DirtySetter("")
    # location = DirtySetter('')
    indexer = DirtySetter(1)

    def __init__(self, show, season, episode, ep_file=""):
        self.season = season
        self.episode = episode
        self._location = ep_file

        # setting any of the above sets the dirty flag
        self.dirty = True

        self.show = show
        self.indexer = self.show.indexer

        self.scene_season = 0
        self.scene_episode = 0
        self.scene_absolute_number = 0

        self.lock = threading.Lock()

        self.specifyEpisode(self.season, self.episode)

        self.relatedEps = []

        self.checkForMetaFiles()

        self.wantedQuality = []

    # name = property(lambda self: self._name, dirty_setter("_name"))
    # season = property(lambda self: self._season, dirty_setter("_season"))
    # episode = property(lambda self: self._episode, dirty_setter("_episode"))
    # absolute_number = property(lambda self: self._absolute_number, dirty_setter("_absolute_number"))
    # description = property(lambda self: self._description, dirty_setter("_description"))
    # subtitles = property(lambda self: self._subtitles, dirty_setter("_subtitles"))
    # subtitles_searchcount = property(lambda self: self._subtitles_searchcount, dirty_setter("_subtitles_searchcount"))
    # subtitles_lastsearch = property(lambda self: self._subtitles_lastsearch, dirty_setter("_subtitles_lastsearch"))
    # airdate = property(lambda self: self._airdate, dirty_setter("_airdate"))
    # hasnfo = property(lambda self: self._hasnfo, dirty_setter("_hasnfo"))
    # hastbn = property(lambda self: self._hastbn, dirty_setter("_hastbn"))
    # status = property(lambda self: self._status, dirty_setter("_status"))
    # indexer = property(lambda self: self._indexer, dirty_setter("_indexer"))
    # indexerid = property(lambda self: self._indexerid, dirty_setter("_indexerid"))
    # # location = property(lambda self: self._location, dirty_setter("_location"))
    # file_size = property(lambda self: self._file_size, dirty_setter("_file_size"))
    # release_name = property(lambda self: self._release_name, dirty_setter("_release_name"))
    # is_proper = property(lambda self: self._is_proper, dirty_setter("_is_proper"))
    # version = property(lambda self: self._version, dirty_setter("_version"))
    # release_group = property(lambda self: self._release_group, dirty_setter("_release_group"))

    def _set_location(self, new_location):
        logger.debug(f"Setter sets location to {new_location}")

        # self._location = newLocation
        if self._location != new_location:
            self.dirty = True

        self._location = new_location

        if new_location and os.path.isfile(new_location):
            self.file_size = os.path.getsize(new_location)
        else:
            self.file_size = 0

    location = property(lambda self: self._location, _set_location)

    @property
    def idxr(self):
        return self.show.idxr

    @property
    def indexer_name(self):
        return self.show.indexer_name

    def refreshSubtitles(self):
        """Look for subtitles files and refresh the subtitles property"""
        self.subtitles, save_subtitles = subtitles.refresh_subtitles(self)
        if save_subtitles:
            self.saveToDB()

    def download_subtitles(self, force=False, force_lang=None):
        if not os.path.isfile(self.location):
            logger.debug(
                "{id}: Episode file doesn't exist, can't download subtitles for {ep}".format(id=self.show.indexerid, ep=episode_num(self.season, self.episode))
            )
            return

        if not subtitles.needs_subtitles(self.subtitles, force_lang):
            logger.debug(
                "Episode already has all needed subtitles, skipping episode {ep} of show {show}".format(
                    ep=episode_num(self.season, self.episode), show=self.show.name
                )
            )
            return

        logger.debug(
            "Checking subtitle candidates for {show} {ep} ({location})".format(
                show=self.show.name, ep=episode_num(self.season, self.episode), location=os.path.basename(self.location)
            )
        )

        self.subtitles, new_subtitles = subtitles.download_subtitles(self, force_lang)

        self.subtitles_searchcount += 1 if self.subtitles_searchcount else 1
        self.subtitles_lastsearch = datetime.datetime.now().strftime(dateTimeFormat)
        self.saveToDB()

        if new_subtitles:
            subtitle_list = ", ".join([subtitles.name_from_code(code) for code in new_subtitles])
            logger.debug(
                "{id}: Downloaded {subtitles} subtitles for {show} {ep}".format(
                    id=self.show.indexerid, subtitles=subtitle_list, show=self.show.name, ep=episode_num(self.season, self.episode)
                )
            )

            notifiers.notify_subtitle_download(self.pretty_name(), subtitle_list)
        else:
            logger.debug(
                "{id}: No subtitles downloaded for {show} {ep}".format(id=self.show.indexerid, show=self.show.name, ep=episode_num(self.season, self.episode))
            )

        return new_subtitles

    def checkForMetaFiles(self):

        oldhasnfo = self.hasnfo
        oldhastbn = self.hastbn

        cur_nfo = False
        cur_tbn = False

        # check for nfo and tbn
        if os.path.isfile(self.location):
            for cur_provider in settings.metadata_provider_dict.values():
                if cur_provider.episode_metadata:
                    new_result = cur_provider._has_episode_metadata(self)
                else:
                    new_result = False
                cur_nfo = new_result or cur_nfo

                if cur_provider.episode_thumbnails:
                    new_result = cur_provider._has_episode_thumb(self)
                else:
                    new_result = False
                cur_tbn = new_result or cur_tbn

        self.hasnfo = cur_nfo
        self.hastbn = cur_tbn

        # if either setting has changed return true, if not return false
        return oldhasnfo != self.hasnfo or oldhastbn != self.hastbn

    def specifyEpisode(self, season, episode):

        sql_results = self.loadFromDB(season, episode)

        if not sql_results:
            # only load from NFO if we didn't load from DB
            if os.path.isfile(self.location):
                try:
                    self.loadFromNFO(self.location)
                except NoNFOException:
                    logger.exception(
                        "{id}: There was an error loading the NFO for episode {ep}".format(id=self.show.indexerid, ep=episode_num(season, episode))
                    )

                # if we tried loading it from NFO and didn't find the NFO, try the Indexers
                if not self.hasnfo:
                    try:
                        result = self.loadFromIndexer(season, episode)
                    except EpisodeDeletedException:
                        result = False

                    # if we failed SQL *and* NFO, Indexers then fail
                    if not result:
                        raise EpisodeNotFoundException("Couldn't find episode {ep}".format(ep=episode_num(season, episode)))

    def loadFromDB(self, season, episode):
        # logger.debug("{id}: Loading episode details for {name} {ep} from DB".format
        #            (id=self.show.indexerid, name=self.show.name,
        #             ep=episode_num(season, episode)))

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT * FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?", [self.show.indexerid, season, episode])

        if len(sql_results) > 1:
            raise MultipleEpisodesInDatabaseException("Your DB has two records for the same show somehow.")
        elif not sql_results:
            logger.debug("{id}: Episode {ep} not found in the database".format(id=self.show.indexerid, ep=episode_num(season, episode)))
            return False
        else:
            if sql_results[0]["name"]:
                self.name = sql_results[0]["name"]

            self.season = season
            self.episode = episode
            self.absolute_number = try_int(sql_results[0]["absolute_number"], 0)
            self.description = sql_results[0]["description"]
            if not self.description:
                self.description = ""
            if sql_results[0]["subtitles"] and sql_results[0]["subtitles"]:
                self.subtitles = sql_results[0]["subtitles"].split(",")
            self.subtitles_searchcount = int(sql_results[0]["subtitles_searchcount"])
            self.subtitles_lastsearch = sql_results[0]["subtitles_lastsearch"]
            self.airdate = datetime.date.fromordinal(int(sql_results[0]["airdate"]))
            # logger.debug(f"1 Status changes from {self.status} to {sql_results[0]["status"]}")
            self.status = int(sql_results[0]["status"] or -1)

            # don't overwrite my location
            if sql_results[0]["location"] and not self._location:
                self.location = os.path.normpath(sql_results[0]["location"])

            if sql_results[0]["file_size"]:
                self.file_size = int(sql_results[0]["file_size"])
            else:
                self.file_size = 0

            self.indexerid = int(sql_results[0]["indexerid"])
            self.indexer = int(sql_results[0]["indexer"])

            sickchill.oldbeard.scene_numbering.xem_refresh(self.show.indexerid, self.show.indexer)

            self.scene_season = try_int(sql_results[0]["scene_season"], 0)
            self.scene_episode = try_int(sql_results[0]["scene_episode"], 0)
            self.scene_absolute_number = try_int(sql_results[0]["scene_absolute_number"], 0)

            if self.scene_absolute_number == 0:
                self.scene_absolute_number = sickchill.oldbeard.scene_numbering.get_scene_absolute_numbering(
                    self.show.indexerid, self.show.indexer, self.absolute_number
                )

            if self.scene_season == 0 or self.scene_episode == 0:
                self.scene_season, self.scene_episode = sickchill.oldbeard.scene_numbering.get_scene_numbering(
                    self.show.indexerid, self.show.indexer, self.season, self.episode
                )

            if sql_results[0]["release_name"] is not None:
                self.release_name = sql_results[0]["release_name"]

            if sql_results[0]["is_proper"]:
                self.is_proper = int(sql_results[0]["is_proper"])

            if sql_results[0]["version"]:
                self.version = int(sql_results[0]["version"])

            if sql_results[0]["release_group"] is not None:
                self.release_group = sql_results[0]["release_group"]

            self.dirty = False
            return True

    def loadFromIndexer(self, season=None, episode=None):

        myEp = self.idxr.episode(self.show, season or self.season, episode or self.episode)
        if not myEp:
            if self.name:
                logger.debug("{} timed out but we have enough info from other sources, allowing the error".format(self.indexer_name))
                return
            else:
                logger.exception("{} timed out, unable to create the episode".format(self.indexer_name))
                return False

        if not myEp.get("episodeName"):
            if self.name:
                logger.info(
                    "This episode {show} - {ep} has no name on {indexer}. Keeping the name: {title}".format(
                        show=self.show.name, ep=episode_num(season, episode), indexer=self.indexer_name, title=self.name
                    )
                )
            else:
                logger.info(
                    "This episode {show} - {ep} has no name on {indexer}. Setting to an empty string".format(
                        show=self.show.name, ep=episode_num(season, episode), indexer=self.indexer_name
                    )
                )
        else:
            self.name = myEp.get("episodeName")

        if not myEp.get("absoluteNumber"):
            logger.debug(
                "{id}: This episode {show} - {ep} has no absolute number on {indexer}".format(
                    id=self.show.indexerid, show=self.show.name, ep=episode_num(season, episode), indexer=self.indexer_name
                )
            )
        else:
            logger.debug(
                "{id}: The absolute number for {ep} is: {absolute} ".format(
                    id=self.show.indexerid, ep=episode_num(season or self.season, episode or self.episode), absolute=myEp["absoluteNumber"]
                )
            )
            self.absolute_number = try_int(myEp["absoluteNumber"], 0)

        self.season = (season, self.season)[season is None]
        self.episode = (episode, self.episode)[season is None]

        sickchill.oldbeard.scene_numbering.xem_refresh(self.show.indexerid, self.show.indexer)

        self.scene_absolute_number = sickchill.oldbeard.scene_numbering.get_scene_absolute_numbering(
            self.show.indexerid, self.show.indexer, self.absolute_number
        )

        self.scene_season, self.scene_episode = sickchill.oldbeard.scene_numbering.get_scene_numbering(
            self.show.indexerid, self.show.indexer, self.season, self.episode
        )

        if myEp.get("overview"):
            self.description = myEp.get("overview")

        first_aired = myEp.get("firstAired")
        if not first_aired or first_aired == "0000-00-00":
            if self.status in (WANTED, UNAIRED, UNKNOWN):
                # Allow changing the airdate of wanted/unaired/unknown in cases of postponed episodes
                first_aired = str(datetime.date.min)
            else:
                first_aired = str(self.airdate)

        rawAirdate = [int(x) for x in first_aired.split("-")]

        try:
            self.airdate = datetime.date(rawAirdate[0], rawAirdate[1], rawAirdate[2])
        except (ValueError, IndexError, TypeError):
            # Changed this to error because it should NEVER happen now
            logger.info(
                "Malformed air date of {aired} retrieved from {indexer} for ({show} - {ep})".format(
                    aired=first_aired, indexer=self.indexer_name, show=self.show.name, ep=episode_num(season, episode)
                ),
                logger.ERROR,
            )
            # if I'm incomplete on the indexer but I once was complete then just delete myself from the DB for now
            if self.indexerid != -1:
                self.deleteEpisode()
            return False

        if myEp.get("id"):
            self.indexerid = myEp.get("id")

        if not self.indexerid:
            logger.exception("Failed to retrieve ID from {indexer}".format(indexer=self.indexer_name))
            if self.indexerid != -1:
                self.deleteEpisode()
            return False

        # don't update show status if show dir is missing, unless it's missing on purpose
        if not os.path.isdir(self.show._location) and not settings.CREATE_MISSING_SHOW_DIRS and not settings.ADD_SHOWS_WO_DIR:
            logger.info("The show dir {0} is missing, not bothering to change the episode statuses since it'd probably be invalid".format(self.show._location))
            return

        if self.location:
            logger.debug(
                "{id}: Setting status for {ep} based on status {status} and location {location}".format(
                    id=self.show.indexerid, ep=episode_num(season, episode), status=statusStrings[self.status], location=self.location
                )
            )

        if not os.path.isfile(self.location):
            if self.airdate >= datetime.date.today() or self.airdate <= datetime.date.min:
                logger.debug("{0}: Episode airs in the future or has no airdate, marking it {1}".format(self.show.indexerid, statusStrings[UNAIRED]))
                self.status = UNAIRED
            elif self.status in [UNAIRED, UNKNOWN]:
                # Only do UNAIRED/UNKNOWN, it could already be snatched/ignored/skipped, or downloaded/archived to disconnected media
                logger.debug("Episode has already aired, marking it {0}".format(statusStrings[self.show.default_ep_status]))
                self.status = self.show.default_ep_status if self.season > 0 else SKIPPED  # auto-skip specials
            else:
                logger.debug("Not touching status [ {0} ] It could be skipped/ignored/snatched/archived".format(statusStrings[self.status]))

        # if we have a media file then it's downloaded
        elif sickchill.oldbeard.helpers.is_media_file(self.location):
            # leave propers alone, you have to either post-process them or manually change them back
            if self.status not in Quality.SNATCHED_PROPER + Quality.DOWNLOADED + Quality.SNATCHED + Quality.ARCHIVED:
                logger.debug(f"5 Status changes from {self.status} to {Quality.statusFromName(self.location)}")
                self.status = Quality.statusFromName(self.location, anime=self.show.is_anime)

        # shouldn't get here probably
        else:
            logger.debug(f"6 Status changes from {self.status} to {UNKNOWN}")
            self.status = UNKNOWN

    def loadFromNFO(self, location):

        if not os.path.isdir(self.show._location):
            logger.info(f"{self.show.indexerid}: The show dir is missing, not bothering to try loading the episode NFO")
            return

        logger.debug(f"{self.show.indexerid}: Loading episode details from the NFO file associated with {location}")

        self.location = location

        if self.location != "":

            if self.status == UNKNOWN and sickchill.oldbeard.helpers.is_media_file(self.location):
                logger.debug(f"7 Status changes from {self.status} to {Quality.statusFromName(self.location, anime=self.show.is_anime)}")
                self.status = Quality.statusFromName(self.location, anime=self.show.is_anime)

            nfoFile = replace_extension(self.location, "nfo")
            logger.debug(f"{self.show.indexerid}: Using NFO name {nfoFile}")

            if os.path.isfile(nfoFile):
                try:
                    showXML = ElementTree.ElementTree(file=nfoFile)
                except (SyntaxError, ValueError) as error:
                    logger.exception(f"Error loading the NFO, backing up the NFO and skipping for now: {error}")
                    try:
                        os.rename(nfoFile, f"{nfoFile}.old")
                    except Exception as error:
                        logger.exception(f"Failed to rename your episode's NFO file - you need to delete it or fix it: {error}")
                    raise NoNFOException("Error in NFO format")

                for epDetails in showXML.iter("episodedetails"):
                    if (
                        epDetails.findtext("season") is None
                        or int(epDetails.findtext("season")) != self.season
                        or epDetails.findtext("episode") is None
                        or int(epDetails.findtext("episode")) != self.episode
                    ):
                        logger.debug(
                            f"{self.show.indexerid}: NFO has an <episodedetails> block for a different episode - wanted {episode_num(self.season, self.episode)} but got {episode_num(epDetails.findtext('season'), epDetails.findtext('episode'))}"
                        )
                        continue

                    if epDetails.findtext("title") is None or epDetails.findtext("aired") is None:
                        raise NoNFOException("Error in NFO format (missing episode title or airdate)")

                    self.name = epDetails.findtext("title")
                    self.episode = int(epDetails.findtext("episode"))
                    self.season = int(epDetails.findtext("season"))

                    sickchill.oldbeard.scene_numbering.xem_refresh(self.show.indexerid, self.show.indexer)

                    self.scene_absolute_number = sickchill.oldbeard.scene_numbering.get_scene_absolute_numbering(
                        self.show.indexerid, self.show.indexer, self.absolute_number
                    )

                    self.scene_season, self.scene_episode = sickchill.oldbeard.scene_numbering.get_scene_numbering(
                        self.show.indexerid, self.show.indexer, self.season, self.episode
                    )

                    self.description = epDetails.findtext("plot")
                    if self.description is None:
                        self.description = ""

                    if epDetails.findtext("aired"):
                        rawAirdate = [int(x) for x in epDetails.findtext("aired").split("-")]
                        self.airdate = datetime.date(rawAirdate[0], rawAirdate[1], rawAirdate[2])

                    self.hasnfo = True
            else:
                self.hasnfo = False

            if os.path.isfile(replace_extension(nfoFile, "tbn")):
                self.hastbn = True
            else:
                self.hastbn = False

    def __str__(self):

        return "\n".join(
            [
                f"{self.show.name} - S{self.season}E{self.episode} - {self.name}",
                f"location: {self.location}",
                f"description: {self.description}",
                f"subtitles: {','.join(self.subtitles)}",
                f"subtitles_searchcount: {self.subtitles_searchcount}",
                f"subtitles_lastsearch: {self.subtitles_lastsearch}",
                f"airdate: {self.airdate.toordinal()} ({self.airdate})",
                f"hasnfo: {self.hasnfo}",
                f"hastbn: {self.hastbn}",
                f"status: {self.status}",
            ]
        )

    def createMetaFiles(self):

        if not os.path.isdir(self.show._location):
            logger.info(f"{self.show.indexerid}: The show dir is missing, not bothering to try to create metadata")
            return

        self.createNFO()
        self.createThumbnail()

        if self.checkForMetaFiles():
            self.saveToDB()

    def createNFO(self):

        result = False

        for cur_provider in settings.metadata_provider_dict.values():
            result = cur_provider.create_episode_metadata(self) or result
            result = cur_provider.update_episode_metadata(self) or result

        return result

    def createThumbnail(self):

        result = False

        for cur_provider in settings.metadata_provider_dict.values():
            result = cur_provider.create_episode_thumb(self) or result

        return result

    def deleteEpisode(self):

        logger.debug(_("Deleting {show} {ep} from the DB").format(show=self.show.name, ep=episode_num(self.season, self.episode)))

        # remove myself from the show dictionary
        if self.show.getEpisode(self.season, self.episode, noCreate=True) == self:
            logger.debug(_("Removing myself from my show's list"))
            del self.show.episodes[self.season][self.episode]

        # delete myself from the DB
        logger.debug(_("Deleting myself from the database"))
        main_db_con = db.DBConnection()
        main_db_con.action("DELETE FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?", [self.show.indexerid, self.season, self.episode])
        raise EpisodeDeletedException()

    def get_sql(self, forceSave=False):
        """
        Creates SQL queue for this episode if any of its data has been changed since the last save.

        forceSave: If True it will create SQL queue even if no data has been changed since the
                    last save (aka if the record is not dirty).
        """
        try:
            if not self.dirty and not forceSave:
                logger.debug(f"{self.show.indexerid}: Not creating SQL queue - record is not dirty")
                return

            main_db_con = db.DBConnection()
            rows = main_db_con.select(
                "SELECT episode_id, subtitles FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?",
                [self.show.indexerid, self.season, self.episode],
            )

            epID = None
            if rows:
                epID = int(rows[0]["episode_id"])

            if epID:
                # use a custom update method to get the data into the DB for existing records.
                # Multi or added subtitle or removed subtitles
                if settings.SUBTITLES_MULTI or not rows[0]["subtitles"] or not self.subtitles:
                    return [
                        "UPDATE tv_episodes SET indexerid = ?, indexer = ?, name = ?, description = ?, subtitles = ?, "
                        "subtitles_searchcount = ?, subtitles_lastsearch = ?, airdate = ?, hasnfo = ?, hastbn = ?, status = ?, "
                        "location = ?, file_size = ?, release_name = ?, is_proper = ?, showid = ?, season = ?, episode = ?, "
                        "absolute_number = ?, version = ?, release_group = ? WHERE episode_id = ?",
                        [
                            self.indexerid,
                            self.indexer,
                            self.name,
                            self.description,
                            ",".join(self.subtitles),
                            self.subtitles_searchcount,
                            self.subtitles_lastsearch,
                            self.airdate.toordinal(),
                            self.hasnfo,
                            self.hastbn,
                            self.status,
                            self.location,
                            self.file_size,
                            self.release_name,
                            self.is_proper,
                            self.show.indexerid,
                            self.season,
                            self.episode,
                            self.absolute_number,
                            self.version,
                            self.release_group,
                            epID,
                        ],
                    ]
                else:
                    # Don't update the subtitle language when the srt file doesn't contain the alpha2 code, keep value from subliminal
                    return [
                        "UPDATE tv_episodes SET indexerid = ?, indexer = ?, name = ?, description = ?, "
                        "subtitles_searchcount = ?, subtitles_lastsearch = ?, airdate = ?, hasnfo = ?, hastbn = ?, status = ?, "
                        "location = ?, file_size = ?, release_name = ?, is_proper = ?, showid = ?, season = ?, episode = ?, "
                        "absolute_number = ?, version = ?, release_group = ? WHERE episode_id = ?",
                        [
                            self.indexerid,
                            self.indexer,
                            self.name,
                            self.description,
                            self.subtitles_searchcount,
                            self.subtitles_lastsearch,
                            self.airdate.toordinal(),
                            self.hasnfo,
                            self.hastbn,
                            self.status,
                            self.location,
                            self.file_size,
                            self.release_name,
                            self.is_proper,
                            self.show.indexerid,
                            self.season,
                            self.episode,
                            self.absolute_number,
                            self.version,
                            self.release_group,
                            epID,
                        ],
                    ]
            else:
                # use a custom insert method to get the data into the DB.
                return [
                    "INSERT OR IGNORE INTO tv_episodes (episode_id, indexerid, indexer, name, description, subtitles, "
                    "subtitles_searchcount, subtitles_lastsearch, airdate, hasnfo, hastbn, status, location, file_size, "
                    "release_name, is_proper, showid, season, episode, absolute_number, version, release_group) VALUES "
                    "((SELECT episode_id FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?)"
                    ",?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);",
                    [
                        self.show.indexerid,
                        self.season,
                        self.episode,
                        self.indexerid,
                        self.indexer,
                        self.name,
                        self.description,
                        ",".join(self.subtitles),
                        self.subtitles_searchcount,
                        self.subtitles_lastsearch,
                        self.airdate.toordinal(),
                        self.hasnfo,
                        self.hastbn,
                        self.status,
                        self.location,
                        self.file_size,
                        self.release_name,
                        self.is_proper,
                        self.show.indexerid,
                        self.season,
                        self.episode,
                        self.absolute_number,
                        self.version,
                        self.release_group,
                    ],
                ]
        except Exception as error:
            logger.exception("Error while updating database: {0}".format(error))

    def saveToDB(self, forceSave=False):
        """
        Saves this episode to the database if any of its data has been changed since the last save.

        forceSave: If True it will save to the database even if no data has been changed since the
                    last save (aka if the record is not dirty).
        """

        if not self.dirty and not forceSave:
            # logger.debug(f"{self.show.indexerid}: Not saving episode to db - record is not dirty")
            return

        newValueDict = {
            "indexerid": self.indexerid,
            "indexer": self.indexer,
            "name": self.name,
            "description": self.description,
            "subtitles": ",".join(self.subtitles),
            "subtitles_searchcount": self.subtitles_searchcount,
            "subtitles_lastsearch": self.subtitles_lastsearch,
            "airdate": self.airdate.toordinal(),
            "hasnfo": self.hasnfo,
            "hastbn": self.hastbn,
            "status": self.status,
            "location": self.location,
            "file_size": self.file_size,
            "release_name": self.release_name,
            "is_proper": self.is_proper,
            "absolute_number": self.absolute_number,
            "version": self.version,
            "release_group": self.release_group,
        }

        controlValueDict = {"showid": self.show.indexerid, "season": self.season, "episode": self.episode}

        # logger.debug(f"{self.show.indexerid}: Saving episode details to database S{self.season}E{self.episode}: {statusStrings[self.status]}")

        # use a custom update/insert method to get the data into the DB
        main_db_con = db.DBConnection()
        main_db_con.upsert("tv_episodes", newValueDict, controlValueDict)

    def fullPath(self):
        if self.location is None or self.location == "":
            return None
        else:
            return os.path.join(self.show.location, self.location)

    def createStrings(self, pattern=None):
        patterns = ["%S.N.S%SE%0E", "%S.N.S%0SE%E", "%S.N.S%SE%E", "%S.N.S%0SE%0E", "%SN S%SE%0E", "%SN S%0SE%E", "%SN S%SE%E", "%SN S%0SE%0E"]

        strings = []
        if not pattern:
            for p in patterns:
                strings += [self._format_pattern(p)]
            return strings
        return self._format_pattern(pattern)

    def pretty_name(self):
        """
        Returns the name of this episode in a "pretty" human-readable format. Used for logging
        and notifications and such.

        Returns: A string representing the episode's name and season/ep numbers
        """

        if self.show.anime and not self.show.scene:
            return self._format_pattern("%SN - %AB - %EN")
        elif self.show.air_by_date:
            return self._format_pattern("%SN - %AD - %EN")

        return self._format_pattern("%SN - S%0SE%0E - %EN")

    def _ep_name(self):
        """
        Returns the name of the episode to use during renaming. Combines the names of related episodes.
        Eg. "Ep Name (1)" and "Ep Name (2)" becomes "Ep Name"
            "Ep Name" and "Other Ep Name" becomes "Ep Name & Other Ep Name"
        """

        multiNameRegex = r"(.*) \(\d{1,2}\)"

        self.relatedEps = sorted(self.relatedEps, key=lambda x: x.episode)

        if not self.relatedEps:
            goodName = self.name
        else:
            singleName = True
            curGoodName = None

            for curName in [self.name] + [x.name for x in self.relatedEps]:
                match = re.match(multiNameRegex, curName)
                if not match:
                    singleName = False
                    break

                if curGoodName is None:
                    curGoodName = match.group(1)
                elif curGoodName != match.group(1):
                    singleName = False
                    break

            if singleName:
                goodName = curGoodName
            else:
                goodName = self.name
                for relEp in self.relatedEps:
                    goodName += f" & {relEp.name}"

        return goodName

    def _replace_map(self):
        """
        Generates a replacement map for this episode which maps all possible custom naming patterns to the correct
        value for this episode.

        Returns: A dict with patterns as the keys and their replacement values as the values.
        """

        ep_name = self._ep_name()

        def dot(name):
            # assert isinstance(name, unicode), f'{name} is not unicode'
            return helpers.sanitizeSceneName(name)

        def us(name):
            return re.sub("[ -]", "_", name)

        def release_name(name):
            if name:
                name = helpers.remove_non_release_groups(remove_extension(name))
            return name

        def release_group(show, name):
            if name:
                name = helpers.remove_non_release_groups(remove_extension(name))
            else:
                return ""

            try:
                # guess_result = guessit.guessit(name, dict(single_value=True))
                parse_result = NameParser(name, showObj=show, naming_pattern=True).parse(name)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug("Unable to get parse release_group: {0}".format(error))
                return ""

            if not parse_result.release_group:
                return ""

            # return guess_result.get('release_group', '').strip('.- []{}')
            return parse_result.release_group.strip(".- []{}")

        epStatus_, epQual = Quality.splitCompositeStatus(self.status)

        if settings.NAMING_STRIP_YEAR:
            show_name = re.sub(r"\(\d+\)$", "", self.show.name).strip()
        else:
            show_name = self.show.name

        # try to get the release group
        rel_grp = {"SICKCHILL": "SICKCHILL"}
        if hasattr(self, "location"):  # from the location name
            rel_grp["location"] = release_group(self.show, self.location)
            if not rel_grp["location"]:
                del rel_grp["location"]
        if hasattr(self, "_release_group"):  # from the release group field in db
            rel_grp["database"] = self._release_group.strip(".- []{}")
            if not rel_grp["database"]:
                del rel_grp["database"]
        if hasattr(self, "release_name"):  # from the release name field in db
            rel_grp["release_name"] = release_group(self.show, self.release_name)
            if not rel_grp["release_name"]:
                del rel_grp["release_name"]

        # use release_group, release_name, location in that order
        if "database" in rel_grp:
            relgrp = "database"
        elif "release_name" in rel_grp:
            relgrp = "release_name"
        elif "location" in rel_grp:
            relgrp = "location"
        else:
            relgrp = "SICKCHILL"

        # try to get the release encoder to comply with scene naming standards
        encoder = Quality.sceneQualityFromName(self.release_name.replace(rel_grp[relgrp], ""), epQual)
        if encoder:
            logger.debug(f"Found codec for '{show_name}: {ep_name}'.")

        return {
            "%SN": show_name,
            "%S.N": dot(show_name),
            "%S_N": us(show_name),
            "%EN": ep_name,
            "%E.N": dot(ep_name),
            "%E_N": us(ep_name),
            "%QN": Quality.qualityStrings[epQual],
            "%Q.N": dot(Quality.qualityStrings[epQual]),
            "%Q_N": us(Quality.qualityStrings[epQual]),
            "%SQN": Quality.sceneQualityStrings[epQual] + encoder,
            "%SQ.N": dot(Quality.sceneQualityStrings[epQual] + encoder),
            "%SQ_N": us(Quality.sceneQualityStrings[epQual] + encoder),
            "%S": "" if self.season is None else str(self.season),
            "%0S": "" if self.season is None else f"{int(self.season):02d}",
            "%E": "" if self.episode is None else str(self.episode),
            "%0E": "" if self.episode is None else f"{int(self.episode):02d}",
            "%XS": "" if self.scene_season is None else str(self.scene_season),
            "%0XS": "" if self.scene_season is None else f"{int(self.scene_season):02d}",
            "%XE": "" if self.scene_episode is None else str(self.scene_episode),
            "%0XE": "" if self.scene_episode is None else f"{int(self.scene_episode):02d}",
            "%AB": "" if self.absolute_number is None else f"{int(self.absolute_number):03d}",
            "%XAB": "" if self.scene_absolute_number is None else f"{int(self.scene_absolute_number):03d}",
            "%RN": release_name(self.release_name),
            "%RG": rel_grp[relgrp],
            "%CRG": rel_grp[relgrp].upper(),
            "%AD": str(self.airdate or "").replace("-", " ") if self.airdate else "",
            "%A.D": str(self.airdate).replace("-", ".") if self.airdate else "",
            "%A_D": us(str(self.airdate)) if self.airdate else "",
            "%A-D": str(self.airdate) if self.airdate else "",
            "%Y": str(self.airdate.year) if self.airdate else "",
            "%M": str(self.airdate.month) if self.airdate else "",
            "%D": str(self.airdate.day) if self.airdate else "",
            "%CY": str(datetime.date.today().year) if self.airdate else "",
            "%CM": str(datetime.date.today().month) if self.airdate else "",
            "%CD": str(datetime.date.today().day) if self.airdate else "",
            "%0M": f"{int(self.airdate.month):02d}" if self.airdate else "",
            "%0D": f"{int(self.airdate.day):02d}" if self.airdate else "",
            "%RT": "PROPER" if self.is_proper else "",
        }

    @staticmethod
    def _format_string(pattern, replace_map):
        """
        Replaces all template strings with the correct value
        """

        result_name = pattern

        # do the replacements
        for cur_replacement in sorted(replace_map, reverse=True):
            result_name = result_name.replace(cur_replacement, sanitize_filename(replace_map[cur_replacement]))
            result_name = result_name.replace(cur_replacement.lower(), sanitize_filename(replace_map[cur_replacement].lower()))

        return result_name

    def _format_pattern(self, pattern=None, multi=None, anime_type=None):
        """
        Manipulates an episode naming pattern and then fills the template in
        """

        if pattern is None:
            pattern = settings.NAMING_PATTERN

        if multi is None:
            multi = settings.NAMING_MULTI_EP

        if settings.NAMING_CUSTOM_ANIME:
            if anime_type is None:
                anime_type = settings.NAMING_ANIME
        else:
            anime_type = 3

        replace_map = self._replace_map()

        result_name = pattern

        # if there's no release group in the db, let the user know we replaced it
        if replace_map["%RG"] and replace_map["%RG"] != "SICKCHILL":
            if not hasattr(self, "_release_group") or not self._release_group:
                logger.debug(f"Episode has no release group, replacing it with '{replace_map['%RG']}'")
                self._release_group = replace_map["%RG"]  # if release_group is not in the db, put it there

        # if there's no release name then replace it with a reasonable facsimile
        if not replace_map["%RN"]:
            if self.show.air_by_date or self.show.sports:
                result_name = result_name.replace("%RN", "%S.N.%A.D.%E.N-" + replace_map["%RG"])
                result_name = result_name.replace("%rn", "%s.n.%A.D.%e.n-" + replace_map["%RG"].lower())

            elif anime_type != 3:
                result_name = result_name.replace("%RN", "%S.N.%AB.%E.N-" + replace_map["%RG"])
                result_name = result_name.replace("%rn", "%s.n.%ab.%e.n-" + replace_map["%RG"].lower())

            else:
                result_name = result_name.replace("%RN", "%S.N.S%0SE%0E.%E.N-" + replace_map["%RG"])
                result_name = result_name.replace("%rn", "%s.n.s%0se%0e.%e.n-" + replace_map["%RG"].lower())

            # logger.debug(f"Episode has no release name, replacing it with a generic one: {result_name}")

        if not replace_map["%RT"]:
            result_name = re.sub("([ _.-]*)%RT([ _.-]*)", r"\2", result_name)

        # split off ep name part only
        name_groups = re.split(r"[\\/]", result_name)

        # figure out the double-ep numbering style for each group, if applicable
        for cur_name_group in name_groups:
            season_ep_regex = r"""
                                (?P<pre_sep>[ _.-]*)
                                ((?:s(?:eason|eries)?\s*)?%0?S(?![._]?N))
                                (.*?)
                                (%0?E(?![._]?N))
                                (?P<post_sep>[ _.-]*)
                              """
            ep_only_regex = r"(E?%0?E(?![._]?N))"

            # try the normal way
            season_ep_match = re.search(season_ep_regex, cur_name_group, re.I | re.X)
            ep_only_match = re.search(ep_only_regex, cur_name_group, re.I | re.X)

            # if we have a season and episode then collect the necessary data
            if season_ep_match:
                season_format = season_ep_match.group(2)
                ep_sep = season_ep_match.group(3)
                ep_format = season_ep_match.group(4)
                sep = season_ep_match.group("pre_sep")
                if not sep:
                    sep = season_ep_match.group("post_sep")
                if not sep:
                    sep = " "

                # force 2-3-4 format if they chose to extend
                if multi in (NAMING_EXTEND, NAMING_LIMITED_EXTEND, NAMING_LIMITED_EXTEND_E_PREFIXED):
                    ep_sep = "-"

                regex_used = season_ep_regex

            # if there's no season then there's not much choice so we'll just force them to use 03-04-05 style
            elif ep_only_match:
                season_format = ""
                ep_sep = "-"
                ep_format = ep_only_match.group(1)
                sep = ""
                regex_used = ep_only_regex

            else:
                continue

            # we need at least this much info to continue
            if not ep_sep or not ep_format:
                continue

            # start with the ep string, eg. E03
            ep_string = self._format_string(ep_format.upper(), replace_map)
            for other_ep in self.relatedEps:

                # for limited extend we only append the last ep
                if multi in (NAMING_LIMITED_EXTEND, NAMING_LIMITED_EXTEND_E_PREFIXED) and other_ep != self.relatedEps[-1]:
                    continue

                elif multi == NAMING_DUPLICATE:
                    # add " - S01"
                    ep_string += sep + season_format

                elif multi == NAMING_SEPARATED_REPEAT:
                    ep_string += sep

                # add "E04"
                ep_string += ep_sep

                if multi == NAMING_LIMITED_EXTEND_E_PREFIXED:
                    ep_string += "E"

                ep_string += other_ep._format_string(ep_format.upper(), other_ep._replace_map())

            if anime_type != 3:
                if self.absolute_number == 0:
                    curAbsolute_number = self.episode
                else:
                    curAbsolute_number = self.absolute_number

                if self.season != 0:  # dont set absolute numbers if we are on specials !
                    if anime_type == 1:  # this crazy person wants both ! (note: +=)
                        ep_string += f"{sep}{curAbsolute_number:03d}"
                    elif anime_type == 2:  # total anime freak only need the absolute number ! (note: =)
                        ep_string = f"{curAbsolute_number:03d}"

                    for relEp in self.relatedEps:
                        if relEp.absolute_number != 0:
                            ep_string += f"-{relEp.absolute_number:03d}"
                        else:
                            ep_string += f"-{relEp.episode:03d}"

            regex_replacement = None
            if anime_type == 2 and not ep_only_match:
                regex_replacement = fr"\g<pre_sep>{ep_string}\g<post_sep>"
            elif season_ep_match:
                regex_replacement = fr"\g<pre_sep>\g<2>\g<3>{ep_string}\g<post_sep>"
            elif ep_only_match:
                regex_replacement = ep_string

            if regex_replacement:
                # fill out the template for this piece and then insert this piece into the actual pattern
                cur_name_group_result = re.sub("(?i)(?x)" + regex_used, regex_replacement, cur_name_group)
                # cur_name_group_result = cur_name_group.replace(ep_format, ep_string)
                result_name = result_name.replace(cur_name_group, cur_name_group_result)

        result_name = self._format_string(result_name, replace_map)

        logger.debug(f"formatting pattern: {pattern} -> {result_name}")

        return result_name

    def proper_path(self):
        """
        Figures out the path where this episode SHOULD live according to the renaming rules, relative from the show dir
        """

        anime_type = settings.NAMING_ANIME
        if not self.show.is_anime:
            anime_type = 3

        result = self.formatted_filename(anime_type=anime_type)

        # if they want us to flatten it and we're allowed to flatten it then we will
        if not (self.show.season_folders or settings.NAMING_FORCE_FOLDERS):
            return result

        # if not we append the folder on and use that
        else:
            result = os.path.join(self.formatted_dir(anime_type=anime_type), result)

        return result

    def formatted_dir(self, pattern=None, multi=None, anime_type=None):
        """
        Just the folder name of the episode
        """

        if pattern is None:
            # we only use ABD if it's enabled, this is an ABD show, AND this is not a multi-ep
            if self.show.air_by_date and settings.NAMING_CUSTOM_ABD and not self.relatedEps:
                pattern = settings.NAMING_ABD_PATTERN
            elif self.show.sports and settings.NAMING_CUSTOM_SPORTS and not self.relatedEps:
                pattern = settings.NAMING_SPORTS_PATTERN
            elif self.show.anime and settings.NAMING_CUSTOM_ANIME:
                pattern = settings.NAMING_ANIME_PATTERN
            else:
                pattern = settings.NAMING_PATTERN

        # split off the dirs only, if they exist
        name_groups = re.split(r"[\\/]", pattern)

        if len(name_groups) == 1:
            return ""
        else:
            return self._format_pattern(os.sep.join(name_groups[:-1]), multi, anime_type)

    def formatted_filename(self, pattern=None, multi=None, anime_type=None):
        """
        Just the filename of the episode, formatted based on the naming settings
        """

        if pattern is None:
            # we only use ABD if it's enabled, this is an ABD show, AND this is not a multi-ep
            if self.show.air_by_date and settings.NAMING_CUSTOM_ABD and not self.relatedEps:
                pattern = settings.NAMING_ABD_PATTERN
            elif self.show.sports and settings.NAMING_CUSTOM_SPORTS and not self.relatedEps:
                pattern = settings.NAMING_SPORTS_PATTERN
            elif self.show.anime and settings.NAMING_CUSTOM_ANIME:
                pattern = settings.NAMING_ANIME_PATTERN
            else:
                pattern = settings.NAMING_PATTERN

        # split off the dirs only, if they exist
        name_groups = re.split(r"[\\/]", pattern)

        return sanitize_filename(self._format_pattern(name_groups[-1], multi, anime_type))

    def rename(self):
        """
        Renames an episode file and all related files to the location and filename as specified
        in the naming settings.
        """

        if not os.path.isfile(self.location):
            logger.warning(f"Can't perform rename on {self.location} when it doesn't exist, skipping")
            return

        proper_path = self.proper_path()
        absolute_proper_path = os.path.join(self.show.location, proper_path)
        absolute_current_path_no_ext, file_ext = os.path.splitext(self.location)
        absolute_current_path_no_ext_length = len(absolute_current_path_no_ext)

        related_subs = []

        current_path = absolute_current_path_no_ext

        if absolute_current_path_no_ext.startswith(self.show.location):
            current_path = absolute_current_path_no_ext[len(self.show.location) :]

        logger.debug(f"Renaming/moving episode from the base path {self.location} to {absolute_proper_path}")

        # if it's already named correctly then don't do anything
        if proper_path == current_path:
            logger.debug(f"{self.indexerid}: File {self.location} is already named correctly, skipping")
            return

        # get related files
        related_files = postProcessor.PostProcessor(self.location).list_associated_files(self.location, subfolders=True, rename=True)

        # get related subs
        if self.show.subtitles and settings.SUBTITLES_DIR:
            # assume that the video file is in the subtitles dir to find associated subs
            subs_path = os.path.join(settings.SUBTITLES_DIR, os.path.basename(self.location))
            related_subs = postProcessor.PostProcessor(self.location).list_associated_files(subs_path, subtitles_only=True, subfolders=True, rename=True)

        logger.debug(f"Files associated to {self.location}: {related_files}")

        # move the ep file
        result = helpers.rename_ep_file(self.location, absolute_proper_path, absolute_current_path_no_ext_length)

        # move related files
        for cur_related_file in related_files:
            # We need to fix something here because related files can be in subfolders and the original code doesn't handle this (at all)
            cur_related_dir = os.path.dirname(os.path.abspath(cur_related_file))
            subfolder = cur_related_dir.replace(os.path.dirname(os.path.abspath(self.location)), "")
            # We now have a subfolder. We need to add that to the absolute_proper_path.
            # First get the absolute proper-path dir
            proper_related_dir = os.path.dirname(os.path.abspath(absolute_proper_path + file_ext))
            proper_related_path = absolute_proper_path.replace(proper_related_dir, proper_related_dir + subfolder)

            cur_result = helpers.rename_ep_file(cur_related_file, proper_related_path, absolute_current_path_no_ext_length + len(subfolder))
            if not cur_result:
                logger.exception(f"{self.indexerid}: Unable to rename file {cur_related_file}")

        for cur_related_sub in related_subs:
            absolute_proper_subs_path = os.path.join(settings.SUBTITLES_DIR, self.formatted_filename())
            cur_result = helpers.rename_ep_file(cur_related_sub, absolute_proper_subs_path, absolute_current_path_no_ext_length)
            if not cur_result:
                logger.exception(f"{self.indexerid}: Unable to rename file {cur_related_sub}")

        # save the ep
        with self.lock:
            if result:
                self.location = absolute_proper_path + file_ext
                for relEp in self.relatedEps:
                    relEp.location = absolute_proper_path + file_ext

        # in case something changed with the metadata just do a quick check
        for curEp in [self] + self.relatedEps:
            curEp.checkForMetaFiles()

        # save any changes to the databas
        sql_l = []
        with self.lock:
            for relEp in [self] + self.relatedEps:
                sql_l.append(relEp.get_sql())

        if sql_l:
            main_db_con = db.DBConnection()
            main_db_con.mass_action(sql_l)

    def airdateModifyStamp(self):
        """
        Make the modify date and time of a file reflect the show air date and time.
        Note: Also called from postProcessor

        """
        if not all([settings.AIRDATE_EPISODES, self.airdate, self.location, self.show, self.show.airs, self.show.network]):
            return

        try:
            airdate_ordinal = self.airdate.toordinal()
            if airdate_ordinal < 1:
                return

            airdatetime = network_timezones.parse_date_time(airdate_ordinal, self.show.airs, self.show.network)

            if settings.FILE_TIMESTAMP_TIMEZONE == "local":
                airdatetime = airdatetime.astimezone(network_timezones.sb_timezone)

            filemtime = datetime.datetime.fromtimestamp(os.path.getmtime(self.location)).replace(tzinfo=network_timezones.sb_timezone)

            if filemtime != airdatetime:
                airdatetime = airdatetime.timetuple()
                logger.debug(
                    "{0}: About to modify date of '{1}' to show air date {2}".format(
                        self.show.indexerid, self.location, time.strftime("%b %d,%Y (%H:%M)", airdatetime)
                    )
                )
                try:
                    if helpers.touchFile(self.location, time.mktime(airdatetime)):
                        logger.info(
                            "{0}: Changed modify date of '{1}' to show air date {2}".format(
                                self.show.indexerid, os.path.basename(self.location), time.strftime("%b %d,%Y (%H:%M)", airdatetime)
                            )
                        )
                    else:
                        logger.warning(
                            "{0}: Unable to modify date of '{1}' to show air date {2}".format(
                                self.show.indexerid, os.path.basename(self.location), time.strftime("%b %d,%Y (%H:%M)", airdatetime)
                            )
                        )
                except Exception:
                    logger.warning(
                        "{0}: Failed to modify date of '{1}' to show air date {2}".format(
                            self.show.indexerid, os.path.basename(self.location), time.strftime("%b %d,%Y (%H:%M)", airdatetime)
                        )
                    )
        except Exception:
            logger.warning("{0}: Failed to modify date of '{1}'".format(self.show.indexerid, os.path.basename(self.location)))

    def cleanup_download_properties(self):
        """
        Clean the properties related with the current download.
        Use only when replacing with a new release or similar tasks.
        """
        self.file_size = 0
        self.is_proper = False
        self.location = ""
        self.release_group = ""
        self.release_name = ""
        self._location = ""
        self._release_group = ""

    def __getstate__(self):
        d = dict(self.__dict__)
        del d["lock"]
        return d

    def __setstate__(self, d):
        d["lock"] = threading.Lock()
        self.__dict__.update(d)



# -*- coding: utf-8 -*-

"""
`tvdbsimple` is a wrapper, written in Python, for TheTVDb.com 
API v2.  By calling the functions available in `tvdbsimple` you can simplify 
your code and easily access a vast amount of tv and cast data. To find 
out more about TheTVDb API, check out the [official api page](https://api.thetvdb.com/swagger/)

Features
--------

- Full API implementation. Supports Search, Series, Episode, Updated, User and Languages methods.
- Updated with the latest JSON API. 
- Fully tested with automated tests and travis.ci.
- Supports Python 2.7, 3.3, 3.4, 3.5, 3.6.
- Easy to access data using Python class attributes.
- Easy to experiment with `tvdbsimple` functions inside the Python interpreter.

Installation
------------

`tvdbsimple` is available on the [Python Package Index](https://pypi.python.org/pypi/tvdbsimple).

You can install `tvdbsimple` using one of the following techniques:

- Use pip:  `pip install tvdbsimple`
- Download the .zip or .tar.gz file from PyPI and install it yourself
- Download the [source from Github](http://github.com/phate89/tvdbsimple) and install it yourself

If you install it yourself, also install [requests](http://www.python-requests.org/en/latest).

API Key
-------
You will need an API key to TheTVDb to access the API.  To obtain a key, follow these steps:

1) Register for and verify an [account](http://thetvdb.com/?tab=register).
2) [Log into](http://thetvdb.com/?tab=login) your account.
3) [Go to this page](http://thetvdb.com/?tab=apiregister) and fill your details to generate a new API key.

Examples
--------
All the functions are fully documented here but you can find several use examples in the [examples page](https://github.com/phate89/tvdbsimple/blob/master/EXAMPLES.rst).

License
-------
The module is distributed with a GPLv3 license, see [LICENSE](https://www.gnu.org/licenses/gpl-3.0.en.html) for more details

copyright (c) 2017 by phate89.
"""

__title__ = 'tvdbsimple'
__version__ = '1.0.6'
__author__ = 'phate89'
__copyright__ = 'Copyright © 2017 phate89'
__license__ = 'GPLv3'

from .base import APIKeyError
from .keys import keys
from .search import Search
from .series import Series, Series_Episodes, Series_Images
from .languages import Languages
from .episode import Episode
from .updates import Updates
from .user import User, User_Ratings

KEYS=keys()
"""
Contains `API_KEY` and `API_TOKEN`.

To use the module you have to set at least the `API_KEY` value (THETVDb api key).
You can also provide an `API_TOKEN` if you already have a valid one stored. If the 
valid token doesn't work anymore the module will try to retrieve a new one
using the `API_KEY` variable
"""



# -*- coding: utf-8 -*-

"""
tmdbsimple.people
~~~~~~~~~~~~~~~~~
This module implements the People and Credits functionality of tmdbsimple.

Created by Celia Oakley on 2013-10-31.

:copyright: (c) 2013-2020 by Celia Oakley
:license: GPLv3, see LICENSE for more details
"""

from .base import TMDB


class People(TMDB):
    """
    People functionality.

    See: https://developers.themoviedb.org/3/people
    """
    BASE_PATH = 'person'
    URLS = {
        'info': '/{id}',
        'changes': '/{id}/changes',
        'movie_credits': '/{id}/movie_credits',
        'tv_credits': '/{id}/tv_credits',
        'combined_credits': '/{id}/combined_credits',
        'external_ids': '/{id}/external_ids',
        'images': '/{id}/images',
        'tagged_images': '/{id}/tagged_images',
        'translations': '/{id}/translations',
        'latest': '/latest',
        'popular': '/popular',
    }

    def __init__(self, id=0):
        super(People, self).__init__()
        self.id = id

    def info(self, **kwargs):
        """
        Get the primary person details by id.

        Supports append_to_response. Read more about this at
        https://developers.themoviedb.org/3/getting-started/append-to-response.

        Args:
            language: (optional) ISO 639-1 code.
            append_to_response: (optional) Append requests within the same
                namespace to the response.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('info')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def changes(self, **kwargs):
        """
        Get the changes for a person. By default only the last 24 hours are returned.

        You can query up to 14 days in a single query by using the start_date
        and end_date query parameters.

        Args:
            start_date: (optional) Filter the results with a start date.
                Expected format is 'YYYY-MM-DD'.
            end_date: (optional) Filter the results with a end date.
                Expected format is 'YYYY-MM-DD'.
            page: (optional) Minimum 1, maximum 1000, default 1.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('changes')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def movie_credits(self, **kwargs):
        """
        Get the movie credits for a person.

        Args:
            language: (optional) ISO 639-1 code.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('movie_credits')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def tv_credits(self, **kwargs):
        """
        Get the TV show credits for a person.

        You can query for some extra details about the credit with the credit
        method.

        Args:
            language: (optional) ISO 639-1 code.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('tv_credits')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def combined_credits(self, **kwargs):
        """
        Get the movie and TV credits together in a single response.

        Args:
            language: (optional) ISO 639-1 code.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('combined_credits')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def external_ids(self, **kwargs):
        """
        Get the external ids for a person. We currently support the following external sources.

        External Sources
            - IMDB ID
            - Facebook
            - Freebase MID
            - Freebase ID
            - Instagram
            - TVRage ID
            - Twitter

        Args:
            language: (optional) ISO 639-1 code.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('external_ids')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def images(self, **kwargs):
        """
        Get the images for a person.

        Args:
            None

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('images')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def tagged_images(self, **kwargs):
        """
        Get the images that this person has been tagged in.

        Args:
            language: (optional) ISO 639-1 code.
            page: (optional) Minimum 1, maximum 1000, default 1.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('tagged_images')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def translations(self, **kwargs):
        """
        Get a list of translations that have been created for a person.

        Args:
            language: (optional) ISO 639-1 code.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_id_path('translations')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def latest(self, **kwargs):
        """
        Get the most newly created person. This is a live response and will
        continuously change.

        Args:
            language: (optional) ISO 639-1 code.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('latest')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def popular(self, **kwargs):
        """
        Get the list of popular people on TMDb. This list updates daily.

        Args:
            language: (optional) ISO 639-1 code.
            page: (optional) Minimum 1, maximum 1000, default 1.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('popular')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response


class Credits(TMDB):
    """
    Credits functionality.

    See: https://developers.themoviedb.org/3/credits
    """
    BASE_PATH = 'credit'
    URLS = {
        'info': '/{credit_id}',
    }

    def __init__(self, credit_id):
        super(Credits, self).__init__()
        self.credit_id = credit_id

    def info(self, **kwargs):
        """
        Get a movie or TV credit details by id.

        Args:
            None

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_credit_id_path('info')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2016-9-26
@author   : ajphukan
@email    : leocreed1994@gmail.com
@Country  : India
@Language : Bengali
'''

LOCALE = [
    [ 'এইমাত্র' , 'একটা সময়'] ,
    [ '%s সেকেন্ড আগে', '%s এর সেকেন্ডের মধ্যে'] ,
    [ '1 মিনিট আগে', '1 মিনিটে'] ,
    [ '%s এর মিনিট আগে', '%s এর মিনিটের মধ্যে'] ,
    [ '1 ঘন্টা আগে', '1 ঘন্টা'] ,
    [ '%s ঘণ্টা আগে', '%s এর ঘন্টার মধ্যে'] ,
    [ '1 দিন আগে', '1 দিনের মধ্যে'] ,
    [ '%s এর দিন আগে', '%s এর দিন'] ,
    [ '1 সপ্তাহ আগে', '1 সপ্তাহের মধ্যে'] ,
    [ '%s এর সপ্তাহ আগে', '%s সপ্তাহের মধ্যে'] ,
    [ '1 মাস আগে', '1 মাসে'] ,
    [ '%s মাস আগে', '%s মাসে'] ,
    [ '1 বছর আগে', '1 বছরের মধ্যে'] ,
    [ '%s বছর আগে', '%s বছরে'],
]



import re

METHOD_TYPE_RETRIEVER = 0
METHOD_TYPE_MODIFIER = 1
METHOD_TYPE_RE = re.compile("\.?set_")
METHOD_VAR_NAME_RE = re.compile("([ptdf]\.|system\.|get\_|is\_|set\_)+([^=]*)")

int_to_bool = lambda x: x in [1, "1"]
check_success = lambda x: x == 0

def valmap(from_list, to_list, arg_index):
    def func(*args):
        args = list(args)
        args[arg_index] = to_list[from_list.index(args[arg_index])]
        return tuple(args)

    return func


class RPCMethod(object):
    def __init__(self, method_names, *args, **kwargs):
        if isinstance(method_names, str):
            method_names = (method_names,)

        self.method_names = tuple(method_names)
        self.method_type = kwargs.get("method_type",
                                      self._identify_method_type())
        self.var_name = kwargs.get("var_name", self._identify_var_name())
        self.pre_processors = kwargs.get("pre_processors", [])
        self.post_processors = kwargs.get("post_processors", [])

        if kwargs.get("boolean", False):
            self.post_processors.append(int_to_bool)


        self.pre_processors = tuple(self.pre_processors)
        self.post_processors = tuple(self.post_processors)

    def get_key(self):
        return self.key

    def get_method_names(self):
        return self.method_names

    def get_var_name(self):
        return self.var_name

    def get_pre_processors(self):
        return tuple(self.pre_processors)

    def get_post_processors(self):
        return tuple(self.post_processors)

    def get_available_method_name(self, available_methods):
        matches = set(self.get_method_names()).intersection(available_methods)
        return matches.pop() if len(matches) > 0 else None

    def is_available(self, available_methods) -> bool:
        return self.get_available_method_name(available_methods) is not None

    def is_retriever(self) -> bool:
        return self._identify_method_type() == METHOD_TYPE_RETRIEVER

    def _identify_method_type(self):
        match = METHOD_TYPE_RE.search(self.method_names[0])
        return METHOD_TYPE_MODIFIER if match else METHOD_TYPE_RETRIEVER

    def _identify_var_name(self):
        match = re.match(METHOD_VAR_NAME_RE, self.method_names[0])
        if match is None:
            raise Exception("Could not id var_name for method: {0}"
                            .format(self.method_names))

        return match.groups(-1)



"""
Test SNI and SSL
"""

import unittest

import certifi
import requests

from sickchill.oldbeard import providers as providers


def generator(_provider):
    """
    Generate tests for each provider

    :param _provider: to generate tests from
    :return: test
    """

    def _connectivity_test():
        """
        Generate tests
        :return: test to run
        """
        if not _provider.url:
            print("{0} has no url set, skipping".format(_provider.name))
            return

        try:
            requests.head(_provider.url, verify=certifi.where(), timeout=10)
        except requests.exceptions.SSLError as error:
            if "certificate verify failed" in str(error):
                print("Cannot verify certificate for {0}".format(_provider.name))
            else:
                print("SSLError on {0}: {1}".format(_provider.name, str(error)))
                raise
        except requests.exceptions.Timeout:
            print("Provider timed out")
        except requests.exceptions.RequestException:
            raise

    return _connectivity_test


class SniTests(unittest.TestCase):
    pass


if __name__ == "__main__":
    print("==================")
    print("STARTING - Provider Connectivity TESTS and SSL/SNI")
    print("==================")
    print("######################################################################")
    # Just checking all providers - we should make this error on non-existent urls.
    for provider in [p for p in providers.makeProviderList()]:
        test_name = "test_{0}".format(provider.name)
        test = generator(provider)
        setattr(SniTests, test_name, test)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(SniTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import sys
from random import shuffle

import sickchill.oldbeard.helpers
from sickchill import settings
from sickchill.oldbeard.providers import (
    abnormal,
    alpharatio,
    archetorrent,
    binsearch,
    bitcannon,
    bjshare,
    btn,
    cpasbien,
    danishbits,
    demonoid,
    elitetorrent,
    eztv,
    filelist,
    gftracker,
    gimmepeers,
    hd4free,
    hdbits,
    hdspace,
    hdtorrents,
    hdtorrents_it,
    horriblesubs,
    hounddawgs,
    ilcorsaronero,
    immortalseed,
    iptorrents,
    kat,
    limetorrents,
    magnetdl,
    morethantv,
    ncore,
    nebulance,
    newpct,
    norbits,
    nyaa,
    omgwtfnzbs,
    pretome,
    rarbg,
    scc,
    scenetime,
    shazbat,
    skytorrents,
    speedcd,
    thepiratebay,
    tntvillage,
    tokyotoshokan,
    torrent9,
    torrent_paradise,
    torrentbytes,
    torrentday,
    torrentleech,
    torrentproject,
    torrentz,
    tvchaosuk,
    xthor,
    yggtorrent,
)

__all__ = [
    "abnormal",
    "alpharatio",
    "archetorrent",
    "binsearch",
    "bitcannon",
    "bjshare",
    "btn",
    "cpasbien",
    "danishbits",
    "demonoid",
    "elitetorrent",
    "eztv",
    "filelist",
    "gftracker",
    "gimmepeers",
    "hd4free",
    "hdbits",
    "hdspace",
    "hdtorrents",
    "hdtorrents_it",
    "horriblesubs",
    "hounddawgs",
    "ilcorsaronero",
    "immortalseed",
    "iptorrents",
    "kat",
    "limetorrents",
    "magnetdl",
    "morethantv",
    "ncore",
    "nebulance",
    "newpct",
    "norbits",
    "nyaa",
    "omgwtfnzbs",
    "pretome",
    "rarbg",
    "scc",
    "scenetime",
    "shazbat",
    "skytorrents",
    "speedcd",
    "thepiratebay",
    "tntvillage",
    "tokyotoshokan",
    "torrent9",
    "torrent_paradise",
    "torrentbytes",
    "torrentday",
    "torrentleech",
    "torrentproject",
    "torrentz",
    "tvchaosuk",
    "xthor",
    "yggtorrent",
]

broken_providers = [
    # 'torrentz', 'yggtorrent'
]


def sortedProviderList(randomize=False):
    initialList = settings.providerList + settings.newznabProviderList + settings.torrentRssProviderList
    providerDict = {x.get_id(): x for x in initialList}

    newList = []

    # add all modules in the priority list, in order
    for curModule in settings.PROVIDER_ORDER:
        if curModule in providerDict:
            newList.append(providerDict[curModule])

    # add all enabled providers first
    for curModule in providerDict:
        if providerDict[curModule] not in newList and providerDict[curModule].is_enabled:
            newList.append(providerDict[curModule])

    # add any modules that are missing from that list
    for curModule in providerDict:
        if providerDict[curModule] not in newList:
            newList.append(providerDict[curModule])

    if randomize:
        shuffle(newList)

    return newList


def makeProviderList():
    # noinspection PyUnresolvedReferences
    return [x.Provider() for x in (getProviderModule(y) for y in __all__ if y not in broken_providers) if x]


def getProviderModule(name):
    name = name.lower()
    prefix = "sickchill.oldbeard.providers."
    if name in __all__ and prefix + name in sys.modules:
        return sys.modules[prefix + name]
    else:
        raise Exception("Can't find " + prefix + name + " in " + "Providers")


def getProviderClass(provider_id):
    providerMatch = [x for x in settings.providerList + settings.newznabProviderList + settings.torrentRssProviderList if x and x.get_id() == provider_id]

    if len(providerMatch) != 1:
        return None
    else:
        return providerMatch[0]


def check_enabled_providers():
    if not settings.DEVELOPER:
        backlog_enabled, daily_enabled = False, False
        for provider in sortedProviderList():
            if provider.is_active:
                if provider.enable_daily and provider.can_daily:
                    daily_enabled = True

                if provider.enable_backlog and provider.can_backlog:
                    backlog_enabled = True

                if backlog_enabled and daily_enabled:
                    break

        if not (daily_enabled and backlog_enabled):
            searches = ((_("daily searches and backlog searches"), _("daily searches"))[backlog_enabled], _("backlog searches"))[daily_enabled]
            formatted_msg = _(
                "No NZB/Torrent providers found or enabled for {searches}.<br/>" 'Please <a href="{web_root}/config/providers/">check your settings</a>.'
            )
            sickchill.oldbeard.helpers.add_site_message(
                formatted_msg.format(searches=searches, web_root=settings.WEB_ROOT), tag="no_providers_enabled", level="danger"
            )
        else:
            sickchill.oldbeard.helpers.remove_site_message(tag="no_providers_enabled")



# -*- coding: utf-8 -*-

"""
This module implements the keys class of tvdbsimple.
"""

import os
class keys():
    """
    The class is needed to hold the api key value and the token value.
    """
    API_KEY = os.environ.get('TVDB_API_KEY', None)
    """
    It's the TheTVDb api key needed for the api access. The developer must fill it.
    """
    API_TOKEN = os.environ.get('TVDB_API_TOKEN', None)
    """
    Contains the TheTVDb token used for authenticating api requests. The developer might provide a valid one.
    
    If a request fails the module will get a new token automatically upating it.
    """


"""
This module provides methods to parse various data
types into and out of binary form. Similar way of
doing things to variable_handlers; we map MIME types
to format-specific classes in a dictionary, and then
have a generic "run" method that directs requests
passed to it to the correct format-specific method.
"""

from __future__ import absolute_import, division
from __future__ import unicode_literals, print_function

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

import json
import json.decoder
from functools import partial

import xmltodict

class DataHandlerMeta(type):

    def __init__(cls, name, bases, dct):
        if not hasattr(cls, 'registry'):
            cls.registry = {}
        else:
            cls.registry.update(**{mimetype: cls for mimetype in dct.get('mimetypes', [dct.get('mimetype')])})
        super(DataHandlerMeta, cls).__init__(name, bases, dct)

DataHandler = DataHandlerMeta(str('DataHandler'), (object,), {})

class XMLParser(DataHandler):
    mimetypes = ['application/xml', 'text/xml']

    @staticmethod
    def dump(python_object, encoding):
        if python_object:
            return xmltodict.unparse(python_object).encode(encoding)

    @staticmethod
    def load(response, encoding):
        return xmltodict.parse(response, encoding=encoding, xml_attribs=True, dict_constructor=dict)

class JSONParser(DataHandler):
    mimetype = 'application/json'

    @staticmethod
    def dump(python_object, encoding):
        if python_object:
            return json.dumps(python_object).encode(encoding)

    @staticmethod
    def load(response, encoding):
        return json.loads(response.decode(encoding))

class HTTPFormEncoder(DataHandler):
    mimetype = 'application/x-www-form-urlencoded'

    @staticmethod
    def dump(python_object, encoding):
        if python_object:
            return urlencode(python_object).encode(encoding)

class PlainText(DataHandler):
    mimetypes = ['text/plain', 'text/html']

    @staticmethod
    def dump(python_object, encoding):
        if python_object:
            return str(python_object).encode(encoding)

    @staticmethod
    def load(response, encoding):
        response = response.decode(encoding)
        try:
            return json.loads(response)
        except ValueError:
            return response
        except json.decoder.JSONDecodeError:
            return response

class Binary(DataHandler):
    mimetypes = ['application/octet-stream']

    @staticmethod
    def dump(python_object, encoding):
        if python_object:
            return python_object

    @staticmethod
    def load(response, encoding):
        return response

def code(action, data, mimetype, encoding='utf-8'):
    if action == 'dump' and hasattr(data, 'read'):
        data = data.read()
    if action == 'dump' and isinstance(data, bytes):
        return getattr(Binary, action)(data, encoding)
    if action == 'load' and not data:
        return None
    if action == 'load' and mimetype not in DataHandler.registry:
        return getattr(Binary, action)(data, encoding)
    if mimetype in DataHandler.registry and getattr(DataHandler.registry[mimetype], action, None):
        return getattr(DataHandler.registry[mimetype], action)(data, encoding)
    else:
        raise Exception('Cannot parse MIME type {}'.format(mimetype))

encode = partial(code, 'dump')
decode = partial(code, 'load')



from pysrt.srttime import SubRipTime
from pysrt.srtitem import SubRipItem
from pysrt.srtfile import SubRipFile
from pysrt.srtexc import Error, InvalidItem, InvalidTimeString
from pysrt.version import VERSION, VERSION_STRING

__all__ = [
    'SubRipFile', 'SubRipItem', 'SubRipFile', 'SUPPORT_UTF_32_LE',
    'SUPPORT_UTF_32_BE', 'InvalidItem', 'InvalidTimeString'
]

ERROR_PASS = SubRipFile.ERROR_PASS
ERROR_LOG = SubRipFile.ERROR_LOG
ERROR_RAISE = SubRipFile.ERROR_RAISE

open = SubRipFile.open
stream = SubRipFile.stream
from_string = SubRipFile.from_string



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
The history module.

:authors: Josh Marshall, Thomas Calmant
:copyright: Copyright 2020, Thomas Calmant
:license: Apache License 2.0
:version: 0.4.1

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Module version
__version_info__ = (0, 4, 1)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class History(object):
    """
    This holds all the response and request objects for a
    session. A server using this should call "clear" after
    each request cycle in order to keep it from clogging
    memory.
    """
    def __init__(self):
        """
        Sets up members
        """
        self.requests = []
        self.responses = []

    def add_response(self, response_obj):
        """
        Adds a response to the history

        :param response_obj: Response content
        """
        self.responses.append(response_obj)

    def add_request(self, request_obj):
        """
        Adds a request to the history

        :param request_obj: A request object
        """
        self.requests.append(request_obj)

    @property
    def request(self):
        """
        Returns the latest stored request or None
        """
        try:
            return self.requests[-1]
        except IndexError:
            return None

    @property
    def response(self):
        """
        Returns the latest stored response or None
        """
        try:
            return self.responses[-1]
        except IndexError:
            return None

    def clear(self):
        """
        Clears the history lists
        """
        del self.requests[:]
        del self.responses[:]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["akkurat nå", "om litt"],
    ["%s sekunder siden", "om %s sekunder"],
    ["1 minutt siden", "om 1 minutt"],
    ["%s minutter siden", "om %s minutter"],
    ["1 time siden", "om 1 time"],
    ["%s timer siden", "om %s timer"],
    ["1 dag siden", "om 1 dag"],
    ["%s dager siden", "om %s dager"],
    ["1 uke siden", "om 1 uke"],
    ["%s uker siden", "om %s uker"],
    ["1 måned siden", "om 1 måned"],
    ["%s måneder siden", "om %s måneder"],
    ["1 år siden", "om 1 år"],
    ["%s år siden", "om %s år"]
]
        


from __future__ import absolute_import

import re
import requests

try:
    import polling
except ImportError:
    raise ImportError(
        "Please install the python module 'polling' via pip or download it from "
        "https://github.com/justiniso/polling/"
    )

from ..exceptions import (
    reCaptchaServiceUnavailable,
    reCaptchaAPIError,
    reCaptchaTimeout,
    reCaptchaParameter,
    reCaptchaBadJobID
)

from . import reCaptcha


class captchaSolver(reCaptcha):

    def __init__(self):
        super(captchaSolver, self).__init__('9kw')
        self.host = 'https://www.9kw.eu/index.cgi'
        self.maxtimeout = 180
        self.session = requests.Session()

    # ------------------------------------------------------------------------------- #

    @staticmethod
    def checkErrorStatus(response):
        if response.status_code in [500, 502]:
            raise reCaptchaServiceUnavailable(
                f'9kw: Server Side Error {response.status_code}'
            )

        error_codes = {
            1: 'No API Key available.',
            2: 'No API key found.',
            3: 'No active API key found.',
            4: 'API Key has been disabled by the operator. ',
            5: 'No user found.',
            6: 'No data found.',
            7: 'Found No ID.',
            8: 'found No captcha.',
            9: 'No image found.',
            10: 'Image size not allowed.',
            11: 'credit is not sufficient.',
            12: 'what was done.',
            13: 'No answer contain.',
            14: 'Captcha already been answered.',
            15: 'Captcha to quickly filed.',
            16: 'JD check active.',
            17: 'Unknown problem.',
            18: 'Found No ID.',
            19: 'Incorrect answer.',
            20: 'Do not timely filed (Incorrect UserID).',
            21: 'Link not allowed.',
            22: 'Prohibited submit.',
            23: 'Entering prohibited.',
            24: 'Too little credit.',
            25: 'No entry found.',
            26: 'No Conditions accepted.',
            27: 'No coupon code found in the database.',
            28: 'Already unused voucher code.',
            29: 'maxTimeout under 60 seconds.',
            30: 'User not found.',
            31: 'An account is not yet 24 hours in system.',
            32: 'An account does not have the full rights.',
            33: 'Plugin needed a update.',
            34: 'No HTTPS allowed.',
            35: 'No HTTP allowed.',
            36: 'Source not allowed.',
            37: 'Transfer denied.',
            38: 'Incorrect answer without space',
            39: 'Incorrect answer with space',
            40: 'Incorrect answer with not only numbers',
            41: 'Incorrect answer with not only A-Z, a-z',
            42: 'Incorrect answer with not only 0-9, A-Z, a-z',
            43: 'Incorrect answer with not only [0-9,- ]',
            44: 'Incorrect answer with not only [0-9A-Za-z,- ]',
            45: 'Incorrect answer with not only coordinates',
            46: 'Incorrect answer with not only multiple coordinates',
            47: 'Incorrect answer with not only data',
            48: 'Incorrect answer with not only rotate number',
            49: 'Incorrect answer with not only text',
            50: 'Incorrect answer with not only text and too short',
            51: 'Incorrect answer with not enough chars',
            52: 'Incorrect answer with too many chars',
            53: 'Incorrect answer without no or yes',
            54: 'Assignment was not found.'
        }

        if response.text.startswith('{'):
            if response.json().get('error'):
                raise reCaptchaAPIError(error_codes.get(int(response.json().get('error'))))
        else:
            error_code = int(re.search(r'^00(?P<error_code>\d+)', response.text).groupdict().get('error_code', 0))
            if error_code:
                raise reCaptchaAPIError(error_codes.get(error_code))

    # ------------------------------------------------------------------------------- #

    def requestJob(self, jobID):
        if not jobID:
            raise reCaptchaBadJobID(
                "9kw: Error bad job id to request reCaptcha against."
            )

        def _checkRequest(response):
            if response.ok and response.json().get('answer') != 'NO DATA':
                return response

            self.checkErrorStatus(response)

            return None

        response = polling.poll(
            lambda: self.session.get(
                self.host,
                params={
                    'apikey': self.api_key,
                    'action': 'usercaptchacorrectdata',
                    'id': jobID,
                    'info': 1,
                    'json': 1
                }
            ),
            check_success=_checkRequest,
            step=10,
            timeout=(self.maxtimeout + 10)
        )

        if response:
            return response.json().get('answer')
        else:
            raise reCaptchaTimeout("9kw: Error failed to solve reCaptcha.")

    # ------------------------------------------------------------------------------- #

    def requestSolve(self, captchaType, url, siteKey):
        def _checkRequest(response):
            if response.ok and response.text.startswith('{') and response.json().get('captchaid'):
                return response

            self.checkErrorStatus(response)

            return None

        captchaMap = {
            'reCaptcha': 'recaptchav2',
            'hCaptcha': 'hcaptcha'
        }

        response = polling.poll(
            lambda: self.session.post(
                self.host,
                data={
                    'apikey': self.api_key,
                    'action': 'usercaptchaupload',
                    'interactive': 1,
                    'file-upload-01': siteKey,
                    'oldsource': captchaMap[captchaType],
                    'pageurl': url,
                    'maxtimeout': self.maxtimeout,
                    'json': 1
                },
                allow_redirects=False
            ),
            check_success=_checkRequest,
            step=5,
            timeout=(self.maxtimeout + 10)
        )

        if response:
            return response.json().get('captchaid')
        else:
            raise reCaptchaBadJobID('9kw: Error no valid job id was returned.')

    # ------------------------------------------------------------------------------- #

    def getCaptchaAnswer(self, captchaType, url, siteKey, reCaptchaParams):
        jobID = None

        if not reCaptchaParams.get('api_key'):
            raise reCaptchaParameter("9kw: Missing api_key parameter.")

        self.api_key = reCaptchaParams.get('api_key')

        if reCaptchaParams.get('maxtimeout'):
            self.maxtimeout = reCaptchaParams.get('maxtimeout')

        if reCaptchaParams.get('proxy'):
            self.session.proxies = reCaptchaParams.get('proxies')

        try:
            jobID = self.requestSolve(captchaType, url, siteKey)
            return self.requestJob(jobID)
        except polling.TimeoutException:
            raise reCaptchaTimeout(
                f"9kw: reCaptcha solve took to long to execute 'captchaid' {jobID}, aborting."
            )

# ------------------------------------------------------------------------------- #


captchaSolver()



import datetime
import re
import time

from sickchill import logger, settings
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.common import cpu_presets
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("Rarbg")

        self.public = True
        self.minseed = 0
        self.ranked = None
        self.sorting = None
        self.minleech = 0
        self.token = None
        self.token_expires = None
        self.supports_movies = True

        # Spec: https://torrentapi.org/apidocs_v2.txt
        self.url = "https://rarbg.to"
        self.urls = {"api": "http://torrentapi.org/pubapi_v2.php"}

        self.proper_strings = ["{{PROPER|REPACK}}"]

        self.cache = tvcache.TVCache(self, min_time=10)  # only poll RARBG every 10 minutes max

    def login(self):
        if self.token and self.token_expires and datetime.datetime.now() < self.token_expires:
            return True

        login_params = {"get_token": "get_token", "format": "json", "app_id": "sickchill"}

        response = self.get_url(self.urls["api"], params=login_params, returns="json")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        self.token = response.get("token")
        self.token_expires = datetime.datetime.now() + datetime.timedelta(minutes=14) if self.token else None
        return self.token is not None

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        search_params = {
            "app_id": "sickchill",
            "min_seeders": try_int(self.minseed),
            "min_leechers": try_int(self.minleech),
            "limit": 100,
            "format": "json_extended",
            "ranked": try_int(self.ranked),
            "token": self.token,
        }

        if ep_obj is not None:
            ep_indexerid = ep_obj.show.indexerid
            ep_indexer = ep_obj.idxr.slug
        else:
            ep_indexerid = None
            ep_indexer = None

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            if mode == "RSS":
                search_params["sort"] = "last"
                search_params["mode"] = "list"
                search_params.pop("search_string", None)
                search_params.pop("search_tvdb", None)
                if settings.movie_list and settings.movie_list.query.count():
                    search_params["category[]"] = ["17", "18"]
                else:
                    search_params["category"] = ["tv"]
            else:
                search_params["category"] = ("tv", "movies")[mode == "Movie"]
                search_params["sort"] = self.sorting if self.sorting else "seeders"
                search_params["mode"] = "search"

                if ep_indexer == "tvdb" and ep_indexerid:
                    search_params["search_tvdb"] = ep_indexerid
                else:
                    search_params.pop("search_tvdb", None)

            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    search_string = re.sub(r"\((\d{4})\)", r"\1", search_string).replace(" ", ".")
                    search_params["search_string"] = search_string
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                time.sleep(cpu_presets[settings.CPU_PRESET] + 2)
                data = self.get_url(self.urls["api"], params=search_params, returns="json")
                if not isinstance(data, dict):
                    logger.debug("No data returned from provider")
                    continue

                error = data.get("error")
                error_code = data.get("error_code")
                # Don't log when {"error":"No results found","error_code":20}
                # List of errors: https://github.com/rarbg/torrentapi/issues/1#issuecomment-114763312
                if error:
                    if try_int(error_code) != 20:
                        logger.info(error)
                    continue

                torrent_results = data.get("torrent_results")
                if not torrent_results:
                    logger.debug("Data returned from provider does not contain any torrents")
                    continue

                for item in torrent_results:
                    try:
                        title = item.pop("title")
                        download_url = item.pop("download")
                        if not all([title, download_url]):
                            continue

                        seeders = item.pop("seeders")
                        leechers = item.pop("leechers")
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    "Discarding torrent because it doesn't meet the"
                                    " minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                                )
                            continue

                        torrent_size = item.pop("size", -1)
                        size = convert_size(torrent_size) or -1
                        torrent_hash = self.hash_from_magnet(download_url)

                        if mode != "RSS":
                            logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                        result = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": torrent_hash}
                        items.append(result)
                    except Exception as e:
                        logger.info(e)

                    continue

            # For each search mode sort all the items by seeders
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



import os
import sys

import pytest

from .client import DelugeRPCClient, LocalDelugeRPCClient, RemoteException


if sys.version_info > (3,):
    long = int


def client_factory(**kw):
    """Create a disconnected client for test purposes."""
    if sys.platform.startswith('win'):
        auth_path = os.path.join(os.getenv('APPDATA'), 'deluge', 'auth')
    else:
        auth_path = os.path.expanduser("~/.config/deluge/auth")

    with open(auth_path, 'rb') as f:
        filedata = f.read().decode("utf-8").split('\n')[0].split(':')

    username, password = filedata[:2]
    ip = '127.0.0.1'
    port = 58846
    kwargs = {'decode_utf8': True}
    if kw:
        kwargs.update(kw)        
    client = DelugeRPCClient(ip, port, username, password, **kwargs)
    return client


@pytest.fixture
def client(request):
    client = client_factory(**getattr(request, 'param', {}))
    client.connect()
    yield client

    try:
        client.disconnect()
    except:
        pass


def test_connect(client):
    assert client.connected


def test_call_method(client):
    assert isinstance(client.call('core.get_free_space'), (int, long))


def test_call_method_arguments(client):
    assert isinstance(client.call('core.get_free_space', '/'), (int, long))


@pytest.mark.parametrize('client',
                         [{'decode_utf8': True}, {'decode_utf8': False}],
                         ids=['decode_utf8_on', 'decode_utf8_off'],
                         indirect=True)
def test_call_method_exception(client):
    with pytest.raises(RemoteException) as ex_info:
        client.call('core.get_free_space', '1', '2')
    assert ('takes at most 2 arguments' in str(ex_info.value) or
            'takes from 1 to 2 positional arguments' in str(ex_info.value))  # deluge 2.0


def test_attr_caller(client):
    assert isinstance(client.core.get_free_space(), (int, long))
    assert isinstance(client.core.get_free_space('/'), (int, long))


def test_call_method_context_manager():
    with client_factory() as client:
        assert isinstance(client.call('core.get_free_space'), (int, long))


def test_local_client_connect():
    with LocalDelugeRPCClient() as local_client:
        assert local_client.connected


def test_local_client_method():
    with LocalDelugeRPCClient() as local_client:
            assert isinstance(local_client.call('core.get_free_space'), (int, long))



from .common import PageTemplate
from .home import Home
from .routes import Route


@Route("/IRC(/?.*)", name="irc")
class HomeIRC(Home):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def index(self):
        t = PageTemplate(rh=self, filename="IRC.mako")
        return t.render(topmenu="system", header=_("IRC"), title=_("IRC"), controller="IRC", action="index")



#
# Copyright (c) 2013 Dean Gardiner, <gardiner91@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import xmlrpc.client
from base64 import encodebytes


class BasicAuthTransport(xmlrpc.client.Transport):
    def __init__(self, username=None, password=None):
        xmlrpc.client.Transport.__init__(self)

        self.username = username
        self.password = password
        self.verbose = False

    def send_auth(self, h):
        if self.username is not None and self.password is not None:
            h.putheader('AUTHORIZATION', "Basic {}".format(encodebytes("{}:{}".format(self.username, self.password)).replace("\012", "")))

    def single_request(self, host, handler, request_body, verbose=0):
        # issue XML-RPC request
        try:
            connection = self.make_connection(host)
            headers = self._extra_headers[:]
            if verbose:
                connection.set_debuglevel(1)
            if self.accept_gzip_encoding :
                connection.putrequest("POST", handler, skip_accept_encoding=True)
                headers.append(("Accept-Encoding", "gzip"))
            else:
                connection.putrequest("POST", handler)
            headers.append(("Content-Type", "text/xml"))
            headers.append(("User-Agent", self.user_agent))
            self.send_headers(connection, headers)
            self.send_auth(connection)
            self.send_content(connection, request_body)

            response = connection.getresponse(buffering=True)
            if response.status == 200:
                self.verbose = verbose
                return self.parse_response(response)
        except xmlrpc.client.Fault:
            raise
        except Exception:
            self.close()
            raise

        #discard any response data and raise exception
        if response.getheader("content-length", 0):
            response.read()
        raise xmlrpc.client.ProtocolError(
            host + handler,
            response.status, response.reason,
            response.msg,
        )
